from brainy.pipes.base import BrainyPipe
from brainy.process.code import (BashCodeProcess, MatlabCodeProcess,
                                 PythonCodeProcess)
from brainy.process.decorator import (format_with_params,
                                      require_key_in_description)


class CustomPipe(BrainyPipe):
    '''
    CustomPipe is a stub to supervise running of custom user code processes.
    '''


class Submittable(object):
    '''
    Run a piece of custom code given by user in self.description[submit_call].
    '''

    @property
    @format_with_params
    @require_key_in_description
    def submit_call(self):
        'Main code to be submitted.'


class BashCall(BashCodeProcess, Submittable):
    '''Bake and call bash as a single job.'''

    def get_bash_code(self):
        return self.submit_call


class MatlabCall(MatlabCodeProcess, Submittable):
    '''Bake and call python as a single job.'''

    def get_matlab_code(self):
        return self.submit_call


class PythonCall(PythonCodeProcess, Submittable):
    '''Bake and call python as a single job.'''

    def get_python_code(self):
        return self.submit_call


# class MapPython(JsonProcess):
#
#     @property
#     def map(self):
#         bake_code = getattr(self, 'bake_%s_code' % self.code_language)
#         script = bake_code(self.map_call)
#         return json.loads(invoke(script))
#
# class Reduce(object):
#     pass



from brainy.utils import invoke
from brainy.process.base import BrainyProcess
from brainy.project.report import BrainyReporter


class CanCheckData(object):

    def has_data(self):
        '''
        Optionally call the method responsible for checking consistency of the
        data. Code is baked into a script and invoked in the shell. Method will
        interpret any output as error.
        '''
        if 'check_data_call' not in self.description:
            return
        bake_code = getattr(self, 'bake_%s_code' % self.code_language)
        script = bake_code(self.description['check_data_call'])
        (stdoutdata, stderrdata) = invoke(script)
        any_output = (stdoutdata + stderrdata).strip()
        if len(any_output) > 0:
            # Interpret any output as error.
            BrainyReporter.append_warning(
                message='Checking data consistency failed',
                output=any_output,
            )
        return True


class CodeProcess(BrainyProcess, CanCheckData):

    def __init__(self, code_language):
        BrainyProcess.__init__(self)
        self.code_language = code_language

    def submit(self):
        '''Default method for code submission'''
        submit_code_job = getattr(self, 'submit_%s_job' % self.code_language)
        get_code = getattr(self, 'get_%s_code' % self.code_language)

        submission_result = submit_code_job(get_code())
        BrainyReporter.append_message(
            message='Submitting new job',
            output=submission_result
        )

        self.set_flag('submitted')

    def resubmit(self):
        submit_code_job = getattr(self, 'submit_%s_job' % self.code_language)
        get_code = getattr(self, 'get_%s_code' % self.code_language)

        resubmission_result = submit_code_job(get_code(), is_resubmitting=True)
        BrainyReporter.append_message(
            message='Resubmitting job',
            output=resubmission_result
        )

        self.set_flag('resubmitted')
        BrainyProcess.resubmit(self)


class BashCodeProcess(CodeProcess):

    def __init__(self):
        super(BashCodeProcess, self).__init__('bash')


class MatlabCodeProcess(CodeProcess):

    def __init__(self):
        CodeProcess.__init__(self, 'matlab')
        super(MatlabCodeProcess, self).__init__('matlab')


class PythonCodeProcess(CodeProcess):

    def __init__(self):
        super(PythonCodeProcess, self).__init__('python')



'''
brainy
======

A nimble workflow managing tool, at the core of iBRAIN functionality. It
allows creation of projects according to the expected framework layout.
It also oversees the execution of the projects and provides monitoring of any
relevant progress of the conducted computation.

@author: Yauhen Yakimovich <yauhen.yakimovich@uzh.ch>,
         Pelkmans lab <https://www.pelkmanslab.org>

@license: The MIT License (MIT)

Copyright (c) 2014 Pelkmans Lab

'''
from brainy.version import __version__, brainy_version


# Mapping command shortcuts to brainy bin scripts.
DUTIES = {
    'brainy-project': {
        'imperatives': [
            'help',
            'create',
            'restart',
            'run',
            'clean',
            'init',
        ],
        'subjectives': [
            'project',
            'projects',
        ],
        'description': 'Manage brainy projects: create new onces, '
                       'run existing.',
    },
    'brainy-config': {
        'imperatives': [
            'help',
            'init',
            'test',
        ],
        'subjectives': [
            'config',
        ],
        'description': 'Init brainy configuration in ~/.brainy/config.',
    },
    'brainy-web': {
        'imperatives': [
            'help',
            'serve',
        ],
        'subjectives': [
            'web',
            'ui',
            'http',
        ],
        'description': 'Starts a local web server and opens a GUI in browser.',
        'use_exec': True,
    },
}



from __future__ import print_function
import os
import sys
import time
import logging
import webbrowser
from twisted.web.server import Site
from twisted.web.static import File
from twisted.internet import reactor
from twisted.python import log


logger = logging.getLogger(__name__)


class NoCacheWrapper(File):

    def render(self, request):
        request.setHeader('Cache-Control', 'no-cache, no-store, '
                          'must-revalidate')
        request.setHeader('Pragma', 'no-cache')
        request.setHeader('Expires', '0')
        request.setLastModified(time.time())
        return super(NoCacheWrapper, self).render(request)


def serve_brainy_project(brainy_project, port):
    log.startLogging(sys.stdout)
    serving_path = os.path.join(brainy_project.report_folder_path, 'html')
    logger.info('Going to serve path: %s' % serving_path)
    root = NoCacheWrapper(serving_path)
    root.putChild("reports", NoCacheWrapper(brainy_project.report_folder_path))
    factory = Site(root)
    url = 'http://localhost:%d/' % port
    # Launch web browser in a fork
    pid = os.fork()
    if pid == 0:
        webbrowser.open(url, new=0, autoraise=True)
    else:
        print('Start serving a brainy project at %s' % url)
        print('You should be able to see it in your browser in a moment..')
        reactor.listenTCP(port, factory)
        reactor.run()



import os
import shutil
import logging
logger = logging.getLogger(__name__)


WORKFLOW_PATH = os.path.join(os.path.dirname(__file__))

WORKFLOWS = {
    'canonical': [
        # iBRAIN Phase 0
        'raw_input.br',
        'illum_corr.br',
        # iBRAIN Phase 1
        'cell_profiler.br',
        # iBRAIN Phase 2
        'post_analysis.br',
    ],
    'demo': [
        'hello_world.br',
    ],
    'empty': [],
}


def bootstrap_workflow(project_path, workflow_name):
    logger.info('Bootstraping project with an iBRAIN workflow: {%s}' %
                workflow_name)
    if not workflow_name in WORKFLOWS:
        raise Exception('Unknown workflow: %s' % workflow_name)
    for workflow_filename in WORKFLOWS[workflow_name]:
        src_path = os.path.join(WORKFLOW_PATH, workflow_filename)
        dst_path = os.path.join(project_path, workflow_filename)
        logger.info('%s -> %s' % (src_path, dst_path))
        shutil.copy(src_path, dst_path)



import os
import re
from xml.sax.saxutils import escape as escape_xml
from fnmatch import fnmatch, translate as fntranslate
from os.path import basename
from brainy.errors import BrainyProcessError
from brainy.process.base import BrainyProcess
from brainy.pipes.base import BrainyPipe
from brainy.pipes.Tools import get_timestamp_str


class Pipe(BrainyPipe):
    '''
    CellProfiller pipe includes steps like:
     - PreCluster
     - CPCluster
     - CPDataFusion
    '''


class PreCluster(BrainyProcess):
    '''Run CellProfiller with CreateBatchFiles module'''

    def __init__(self):
        super(PreCluster, self).__init__()
        self.__batch_files = None

    @property
    def batch_files(self):
        '''Get batch files for submission with CPCluster'''
        if self.__batch_files is None:
            self.__batch_files = CPCluster.get_batch_files(
                self.list_batch_dir())
        return self.__batch_files

    @property
    def cp_pipeline_fnpattern(self):
        return self.description.get(
            'filename',
            'PreCluster_*.mat',
        )

    def get_cp_pipeline_path(self):
        filename_regex_obj = re.compile(fntranslate(
                                        self.cp_pipeline_fnpattern))
        # Look both in PIPES folder and process folder of the pipe (named
        # after pipe).
        for location in (self.process_path, self.pipes_path):
            cp_pipeline_files = [
                os.path.join(location, filename) for filename
                in os.listdir(location)
                if filename_regex_obj.match(filename)
            ]
            # Check if found anything and complain otherwise.
            if len(cp_pipeline_files) > 1:
                raise BrainyProcessError('More than one CP pipeline settings '
                                         'file found matching: %s in %s' %
                                         (self.cp_pipeline_fnpattern,
                                          self.process_path))
            elif len(cp_pipeline_files) == 1:
                # First found pipeline. Rest is ignored.
                return cp_pipeline_files[0]
        # Failed to find anything.
        raise BrainyProcessError('No CP pipeline settings file'
                                 ' found matching: %s in %s' %
                                 (self.cp_pipeline_fnpattern,
                                  self.process_path))

    def put_on(self):
        super(PreCluster, self).put_on()
        # Make sure those pathnames exists. Create if missing.
        if not os.path.exists(self.batch_path):
            os.makedirs(self.batch_path)
        if not os.path.exists(self.postanalysis_path):
            os.makedirs(self.postanalysis_path)

    def get_matlab_code(self):
        matlab_code = '''
        check_missing_images_in_folder('%(tiff_path)s');
        PreCluster_with_pipeline( ...
            '%(cp_pipeline_file)s','%(tiff_path)s','%(batch_path)s');
        ''' % {
            'cp_pipeline_file': self.get_cp_pipeline_path(),
            'batch_path': self.batch_path,
            'tiff_path': self.tiff_path,
        }
        return matlab_code

    def submit(self):
        matlab_code = self.get_matlab_code()
        submission_result = self.submit_matlab_job(matlab_code)

        print('''
            <status action="%(step_name)s">submitting
            <output>%(submission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'submission_result': escape_xml(submission_result),
        })

        self.set_flag('submitted')

    def resubmit(self):
        matlab_code = self.get_matlab_code()
        resubmission_result = self.submit_matlab_job(matlab_code,
                                                     is_resubmitting=True)

        print('''
            <status action="%(step_name)s">resubmitting
            <output>%(resubmission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'resubmission_result': escape_xml(resubmission_result),
        })

        self.set_flag('resubmitted')
        super(PreCluster, self).resubmit()

    def has_data(self):
        '''Validate if precluster has generated any batch files'''
        # TODO: potentially try to learn the exact expected number of files.
        return len(self.batch_files) > 0


class CPCluster(BrainyProcess):
    '''Submit individual batches of CellProfiller as jobs and check results'''

    def __init__(self):
        super(CPCluster, self).__init__()
        self.__batch_files = None
        self._job_report_exp = 'Batch_\d+_to_\d+.results_\d+'
        self._batch_expr = re.compile('Batch_(\d+_to_\d+)')

    @property
    def reports_path(self):
        # Support old mode
        return self.batch_path

    @staticmethod
    def get_batch_files(all_files):
        '''Match files in the BATCH/* to the pattern of batch file.'''
        batch_expr = re.compile('Batch_\d+_to_\d+.mat')
        return [filename for filename in all_files
                if batch_expr.search(basename(filename))]

    @property
    def batch_files(self):
        '''Get batch files for submission with CPCluster'''
        if self.__batch_files is None:
            self.__batch_files = CPCluster.get_batch_files(
                self.list_batch_dir())
        return self.__batch_files

    def put_on(self):
        super(CPCluster, self).put_on()

    def get_matlab_code(self, batch_filename):
        matlab_code = '''
            CPCluster('%(batchfile)s', '%(clusterfile)s')
        ''' % {
            'batchfile': os.path.join(self.batch_path, 'Batch_data.mat'),
            'clusterfile': batch_filename,
        }
        return matlab_code

    def submit(self):
        results = list()
        for batch_filename in self.batch_files:
            matlab_code = self.get_matlab_code(batch_filename)
            batch_report = batch_filename.replace(
                '.mat', '_%s.results' % get_timestamp_str())
            submission_result = self.submit_matlab_job(
                matlab_code,
                report_file=batch_report,
            )
            results.append(submission_result)

        if not results:
            # We want to explicitly check for errors and report them, if any
            # found.
            self.check_logs_for_errors()
            # If no errors were found, we still have to report a warning!
            raise BrainyProcessError(warning='Failed to find any batches.. '
                                     'check or restart previous step')

        print('''
            <status action="%(step_name)s">submitting (%(batch_count)d) batches..
            <output>%(submission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'batch_count': len(results),
            'submission_result': escape_xml(str(results)),
        })

        self.set_flag('submitted')

    def resubmit(self):
        resubmission_results = list()
        output_batches = self.get_out_files()
        result_batches = self.get_result_files()
        for batch_filename in self.batch_files:
            # Resubmit only those files that have no data, i.e. failed with no
            # output.
            batch_prefix = self.parse_batch(batch_filename)
            if batch_prefix in output_batches \
                    and batch_prefix in result_batches:
                # This batch is complete. Do not resubmit it.
                print '<!-- CP batch %s is complete. Skipping.. -->' % \
                      batch_filename
                continue
            # Batch has not produced any output yet. We can resubmit it.
            matlab_code = self.get_matlab_code(batch_filename)
            batch_report = batch_filename.replace(
                '.mat', '_%s.results' % get_timestamp_str())
            resubmission_result = self.submit_matlab_job(
                matlab_code,
                report_file=batch_report,
                is_resubmitting=True,
            )
            resubmission_results.append(resubmission_result)

        if not resubmission_results:
            # We want to explicitly check for errors and report them, if any
            # found.
            self.check_logs_for_errors()
            # If no errors were found, we still have to report a warning!
            raise BrainyProcessError(warning='Failed to find any batches to '
                                     'resubmit. Check or restart previous '
                                     'step.')

        print('''
            <status action="%(step_name)s">resubmitting (%(batch_count)d) batches..
            <output>%(resubmission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'batch_count': len(resubmission_results),
            'resubmission_result': escape_xml(str(resubmission_results)),
        })

        self.set_flag('resubmitted')
        super(CPCluster, self).resubmit()

    def parse_batch(self, filename):
        '''Return batch_prefix - a span of batch numbers.'''
        return self._batch_expr.search(basename(filename)).group(1)

    def get_out_files(self):
        '''
        Get a list of OUT.mat files from BATCH folder. Each file is produced
        when measurements of a particular CP job are saved, i.e. written to
        disk.
        '''
        return [self.parse_batch(filename) for filename
                in self.list_batch_dir()
                if fnmatch(basename(filename), 'Batch_*_OUT.mat')]

    def get_result_files(self):
        '''Get a list of job report files from BATCH folder.'''
        return [self.parse_batch(filename)
                for filename in self.list_batch_dir()
                if fnmatch(basename(filename), 'Batch_*_*.results')]

    def has_data(self):
        '''Validate the integrity of cpcluster step'''
        output_batches = self.get_out_files()
        # Interested only in unique entries of job report files.
        result_batches = set(self.get_result_files())
        #print('Found %d job results logs vs. %d .MAT output files' %
        #      (len(result_batches), len(output_batches)))

        # At this point the number of expected output has to correspond to the
        # number of batches on the input (that were originally submitted).
        if len(result_batches) != len(output_batches) \
                or len(self.batch_files) != len(output_batches):
            return False

        # Note that sanity check is not required since '*_OUT.mat' are
        # produced right after the actual results were written.

        # So otherwise data is valid.
        return True


class CPDataFusion(BrainyProcess):
    '''
    Submit fusion jobs of measurements obtained from results of CellProfiller
    individual jobs. Fusion results themselves should be validated as well.
    '''

    def __init__(self):
        super(CPDataFusion, self).__init__()
        self.__fused_files = None
        self._job_report_exp = 'DataFusion_.*_\d+\.results'

    @property
    def fused_files(self):
        '''Get fused files that were submitted by CPCluster'''
        if self.__fused_files is None:
            self.__fused_files = list()
            batch_expr = re.compile(
                '^Batch_\d+_to_\d+_(Measurements_.*\.mat)$')
            for filename in self.list_batch_dir():
                match = batch_expr.search(basename(filename))
                if not match:
                    continue
                fused_filename = match.group(1)
                if not fused_filename in self.__fused_files:
                    self.__fused_files.append(fused_filename)
        return self.__fused_files

    def get_matlab_code(self, fused_filename):
        matlab_code = '''
        RunDataFusion( ...
            '%(batch_path)s', '%(fused_filename)s');
        ''' % {
            'batch_path': self.batch_path,
            'fused_filename': fused_filename,
        }
        return matlab_code

    def submit(self):
        results = list()
        for fused_filename in self.fused_files:
            matlab_code = self.get_matlab_code(fused_filename)
            fused_report = 'DataFusion_' + fused_filename.replace(
                '.mat', '_%s.results' % get_timestamp_str())
            submission_result = self.submit_matlab_job(
                matlab_code,
                report_file=fused_report,
            )
            results.append(submission_result)

        if not results:
            raise BrainyProcessError(warning='Failed to find complete '
                                     'measurements.. check or restart previous'
                                     ' step')

        print('''
            <status action="%(step_name)s">submitting (%(results_count)d) fusion jobs..
            <output>%(submission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'results_count': len(results),
            'submission_result': escape_xml(str(results)),
        })

        self.set_flag('submitted')

    def resubmit(self):
        results = list()
        for fused_filename in self.fused_files:
            matlab_code = self.get_matlab_code(fused_filename)
            fused_report = 'DataFusion_' + fused_filename.replace(
                '.mat', '_%s.results' % get_timestamp_str())
            resubmission_result = self.submit_matlab_job(
                matlab_code,
                report_file=fused_report,
                is_resubmitting=True,
            )
            results.append(resubmission_result)

        if not results:
            raise BrainyProcessError(warning='Failed to resubmit data fusion')

        print('''
            <status action="%(step_name)s">resubmitting (%(results_count)d) fusion jobs..
            <output>%(resubmission_result)s</output>
            </status>
        ''' % {
            'step_name': self.step_name,
            'results_count': len(results),
            'resubmission_result': escape_xml(str(results)),
        })

        self.set_flag('resubmitted')
        super(CPDataFusion, self).resubmit()

    @property
    def ignore_measurements_with_substrings(self):
        ignored_names = ['batch_illcor', 'OUT']
        ignored_names += self.description.get(
            'ignore_measurements_with_substrings', [])
        return ignored_names

    def has_data(self):
        '''Validate the integrity of cpfusion step'''
        number_of_expected_fused_files = len(self.fused_files)
        # Find actually present outputs.
        expr = re.compile('Measurements_([^\.]*)\.mat')
        parse_feature = lambda filename: expr.search(basename(filename)) \
            .group(1)
        ignored_names = self.ignore_measurements_with_substrings
        found_files = [parse_feature(filename) for filename
                       in self.list_batch_dir()
                       if fnmatch(basename(filename), 'Measurements_*.mat')
                       and not any([(ignored_name in basename(filename))
                                   for ignored_name in ignored_names])]
        print('Found %d fusion job results logs vs. %d merged .MAT feature '
              'files' % (number_of_expected_fused_files, len(found_files)))
        return number_of_expected_fused_files == len(found_files)



import re
import time
from xml.sax.saxutils import escape as escape_xml_special_chars
from subprocess import (PIPE, Popen)


# http://www.w3.org/TR/REC-xml/#charsets
escape_exp = re.compile(u'[^\u0009\u000a\u000d\u0020-\uD7FF\uE000-\uFFFD]+')


def invoke(command, _in=None):
    '''
    Invoke command as a new system process and return its output.
    '''
    process = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True,
                    executable='/bin/bash')
    if _in is not None:
        return process.communicate(input=_in)
    stdoutdata = process.stdout.read()
    stderrdata = process.stderr.read()
    return (stdoutdata, stderrdata)


def escape_xml(raw_value):
    return escape_exp.sub('', unicode(escape_xml_special_chars(raw_value)))


def merge_dicts(a, b, path=None, overwrite=True):
    "merges b into a"
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                if not overwrite:
                    raise Exception('Conflict at %s' % '.'.join(path +
                                    [str(key)]))
        else:
            a[key] = b[key]
    return a


class Timer(object):
    '''
    timer = Timer()

    with timer:
        # Whatever you want to measure goes here
        time.sleep(2)

    print timer.duration_in_seconds()
    '''
    def __enter__(self):
        self.__start = time.time()

    def __exit__(self, type, value, traceback):
        # Error handling here
        self.__finish = time.time()

    def duration_in_seconds(self):
        return self.__finish - self.__start



import logging
from brainy.utils import invoke
from brainy.scheduler.base import BrainyScheduler
from brainy.errors import BrainyProcessError
logger = logging.getLogger(__name__)


class ShellCommand(BrainyScheduler):
    '''
    "No scheduler" scheme will run commands as serial code. Useful for testing
    and optional fallback to local execution.
    '''

    def submit_job(self, shell_command, queue, report_file):
        # Invoke the shell command.
        (stdoutdata, stderrdata) = invoke(shell_command)
        # Produce a report output.
        with open(report_file, 'w+') as report:
            report.write('--CMD---' + '-' * 80 + '\n')
            report.write(shell_command)
            if len(stdoutdata) > 0:
                report.write('\n-STDOUT-' + '-' * 80 + '\n')
                report.write(stdoutdata)
            if len(stderrdata) > 0:
                report.write('\n-STDERR-' + '-' * 80 + '\n')
                report.write(stderrdata)
        # Fail submission if the child process ended up badly.
        if len(stderrdata) > 0:
            raise BrainyProcessError(
                message='Process job produced some error(s). ',
                output=stderrdata,
                job_report=report_file,
            )
        logger.info('Command was successfully executed: %s', shell_command)
        logger.info('Report file is written to: %s' % report_file)
        return ('Command was successfully executed: "%s"\n' +
                'Report file is written to: %s') % \
               (shell_command, report_file)

    def count_working_jobs(self, key):
        return 0

    def list_jobs(self, states):
        return list()



import os
from sh import touch


class FlagManager(object):

    def _get_flag_prefix(self):
        raise NotImplemented()

    @property
    def is_complete(self):
        return os.path.exists('%s.complete' % self._get_flag_prefix())

    @property
    def is_submitted(self):
        return os.path.exists('%s.submitted' % self._get_flag_prefix())

    @property
    def is_resubmitted(self):
        return os.path.exists('%s.submitted' % self._get_flag_prefix()) \
            and os.path.exists('%s.resubmitted' % self._get_flag_prefix())

    @property
    def has_runlimit(self):
        return os.path.exists('%s.runlimit' % self._get_flag_prefix())

    def reset_submitted(self):
        '''
        If no jobs are found for this project, module, or step, waiting is
        senseless. Remove ".submitted" file and try again.
        '''
        self.reset_flag('submitted')
        # Since we are a high-level call, resetting depending flags guarantees
        # we avoid complicated situations.
        self.reset_resubmitted()

    def reset_resubmitted(self):
        '''
        Reset ".resubmitted" flag assuming some results being already produced
        and the fact that their computation should not be redone. E.g. this
        should be called to resume the computation for "known" errors. Which
        deserve retrying.
        '''
        self.reset_flag('runlimit')
        self.reset_flag('resubmitted')

    def set_flag(self, flag='submitted'):
        flag = '.'.join((self._get_flag_prefix(), flag))
        touch(flag)

    def get_flag(self, flag='submitted'):
        return '%s.%s' % (self._get_flag_prefix(), flag)

    def reset_flag(self, flag='submitted'):
        flag_path = self.get_flag(flag)
        if not os.path.exists(flag_path):
            print('Failed to reset: "%s" flag not found.' % flag)
            return
        os.remove(flag_path)



import os
import yaml
import shutil
from glob import glob
from datetime import datetime
from brainy.log import json_handler
import logging
logger = logging.getLogger(__name__)

# This is a global namespace variable containing a DOM-like structure of
# project report. It is mainly modified during `brainy run project` call.
report_data = {}

UI_WEB_PATH = os.path.join(
    os.path.dirname(os.path.dirname(
                    os.path.dirname(os.path.dirname(__file__)))),
    'ui', 'web')


class BrainyReporter(object):

    @classmethod
    def get_now_str(cls):
        return datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    @classmethod
    def start_report(cls):
        report_data['started_at'] = cls.get_now_str()

    @classmethod
    def finalize_report(cls):
        '''Produce final report structure as JSON'''
        report_data['finished_at'] = cls.get_now_str()
        report_data['log'] = json_handler.houtput.root  # points to dictionary

    @classmethod
    def save_report(cls, report_filepath):
        reports_folder = os.path.dirname(report_filepath)
        if not os.path.exists(reports_folder):
            logger.info('Creating missing reports folder: %s' % reports_folder)
            os.makedirs(reports_folder)
        now_str = datetime.now().strftime('%Y_%m_%d_%H%M%S')
        # Output a human readable YAML file.
        yaml_report_filepath = '%s-report-%s.yaml' % (report_filepath, now_str)
        with open(yaml_report_filepath, 'w+') as yaml_reportfile:
            yaml_reportfile.write(yaml.dump(report_data,
                                  default_flow_style=False))

    @classmethod
    def update_or_generate_static_html(cls, reports_folder_path):
        '''
        Put a copy of static html to browse project reports inside
        'reports/html'.
        '''
        # logger.info('web ui path: %s' % UI_WEB_PATH)
        html_path = os.path.join(reports_folder_path, 'html')
        if not os.path.exists(html_path):
            logger.info('Creating static html to browse reports.')
            # os.makedirs(html_path)
            shutil.copytree(os.path.join(UI_WEB_PATH, 'assets'),
                            os.path.join(html_path, 'assets'))
            shutil.copy(os.path.join(UI_WEB_PATH, 'index.html'),
                        os.path.join(html_path, 'index.html'))
        else:
            logger.info('Updating static html to browse reports.')
        # Generated/update reports.json
        report_list = glob(os.path.join(reports_folder_path,
                           '*-report-*.yaml'))
        report_list = sorted(report_list, reverse=True)
        reports = {
            'modified_at': cls.get_now_str(),
            'reports_list': report_list,
        }
        with open(os.path.join(html_path, 'reports.yaml'), 'w+') \
                as reportsfile:
            reportsfile.write(yaml.dump(reports,
                              default_flow_style=False))

    @classmethod
    def get_current_report_pipe(cls):
        if 'pipes' not in report_data['project']:
            raise Exception('KeyError "pipes". Report data was not properly '
                            'initialized.')
        pipe_index = len(report_data['project']['pipes']) - 1
        if pipe_index < 0:
            raise Exception('Failed to find current pipe. '
                            'Please append a pipe first.')
        return report_data['project']['pipes'][pipe_index]

    @classmethod
    def append_report_pipe(cls, name, **extra):
        pipe = {
            'name': name,
            'processes': [],
        }
        pipe.update(extra)
        if 'pipes' not in report_data['project']:
            report_data['project']['pipes'] = []
        report_data['project']['pipes'].append(pipe)

    @classmethod
    def append_report_process(cls, name, **extra):
        process = {
            'name': name,
        }
        process.update(extra)
        cls.get_current_report_pipe()['processes'].append(process)

    @classmethod
    def get_current_report_step(cls):
        pipe = cls.get_current_report_pipe()
        process_index = len(pipe['processes']) - 1
        if process_index < 0:
            raise Exception('Failed to find current process. '
                            'Please append a process first.')
        return pipe['processes'][process_index]

    @classmethod
    def append_message(cls, message, message_type='info', **kwds):
        process = cls.get_current_report_step()
        if 'messages' not in process:
            process['messages'] = []
        message = {
            'message': message,
            'type': message_type,
        }
        message.update(kwds)
        process['messages'].append(message)

    @classmethod
    def append_warning(cls, message, **kwds):
        '''Unknown error is fatal and should break any further progress'''
        cls.append_message(message, message_type='warning', **kwds)

    @classmethod
    def append_unknown_error(cls, message, **kwds):
        '''Unknown error is fatal and should break any further progress'''
        cls.append_message(message, message_type='unknown_error', **kwds)

    @classmethod
    def append_known_error(cls, message, **kwds):
        '''Known error is typical and causes a series of retries'''
        cls.append_message(message, message_type='known_error', **kwds)



import os
import re
import logging
import pprint
from datetime import datetime
import pipette
from brainy.flags import FlagManager
from brainy.scheduler import SHORT_QUEUE, NORM_QUEUE
from brainy.errors import (UnknownError, KnownError, TermRunLimitError,
                           check_report_file_for_errors, BrainyProcessError)
from brainy.project.report import BrainyReporter
logger = logging.getLogger(__name__)


PROCESS_STATUS = [
    'submitting',
    'waiting',
    'resubmitted',
    'reseted',
    'failed',
    'completed',
]


def format_code(code, lang='bash'):
    result = ''
    if lang == 'python':
        left_strip_width = None
        for line in code.split('\n'):
            if len(line.strip()) == 0:
                continue
            if left_strip_width is None:
                # Strip only first left.
                left_strip_width = 0
                j = 0
                while j < len(line):
                    if line[j] == ' ':
                        left_strip_width += 1
                        j += 1
                    else:
                        break
            result += line.rstrip()[left_strip_width:] + '\n'
    else:
        for line in code.split('\n'):
            if len(line.strip()) == 0:
                continue
            result += line.strip() + '\n'
    return result


class BrainyProcess(pipette.Process, FlagManager):

    def __init__(self):
        super(BrainyProcess, self).__init__()
        self.__reports = None
        self.__batch_listing = None
        self._job_report_exp = None
        # List all the relevant process parameters that can be injected into
        # description values. This provides an additional level of
        # parametrization. Also see format_with_params()
        self.format_parameters = [
            'name',
            'step_name',
            # 'plate_path',  # Deprecated
            'project_path',
            'process_path',
            # 'pipes_path', # Deprecated
            'reports_path',
            # 'batch_path',   # Deprecated
            # 'tiff_path',   # Deprecated
            'data_path',
            'images_path',
            'analysis_path',
            # 'postanalysis_path',  # Deprecated
            # 'jpg_path',  # Deprecated
            'job_submission_queue',
            'job_resubmission_queue',
        ]
        self.compiled_params = dict()

    @property
    def pipes_manager(self):
        # FIXME: Pass everything to parameters. Make process independent from
        # manager.
        return self.parameters['pipes_manager']

    @property
    def config(self):
        return self.parameters['pipes_manager'].config

    @property
    def scheduler(self):
        return self.parameters['pipes_manager'].scheduler

    @property
    def name(self):
        return self.parameters['name']

    @property
    def step_name(self):
        return self.parameters['step_name']

    @property
    def job_report_exp(self):
        return self.parameters.get(
            'job_report_exp',
            '%s_\d+.job_report' % self.name if self._job_report_exp is None
            else self._job_report_exp,
        )

    @property
    def project_path(self):
        # We don't allow to parametrize project_path for security reasons.
        # See method restrict_to_safe_path() below.
        return self.parameters['pipes_manager'].project_path

    def restrict_to_safe_path(self, pathname):
        '''Restrict every relative pathname to point within the plate path'''
        try:
            safe_path = self.project_path
            if not safe_path.endswith('/'):
                safe_path += '/'
            if '../../' in pathname:
                pathname = pathname.replace('../../', safe_path, 1)
            pathname = pathname.replace('..', '')
            # The actual jailing withing the safe_path.
            assert pathname.startswith(safe_path)
            return pathname
        except Exception as error:
            logger.error('Safe check failed: %s' % pathname)
            logger.exception(error)
            raise error

    @property
    def process_path(self):
        return self.restrict_to_safe_path(self.parameters['process_path'])

    @property
    def reports_path(self):
        reports_path = self.parameters.get(
            'reports_path',
            os.path.join(self.process_path, 'job_reports_of_{name}'),
        )
        reports_path = self.format_with_params('reports_path', reports_path)
        return self.restrict_to_safe_path(reports_path)

    @property
    def data_path(self):
        '''All the shared data of the process are located here.'''
        data_path = self.parameters.get(
            'data_path',
            os.path.join(self.process_path, 'data_of_{name}'),
        )
        data_path = self.format_with_params('data_path', data_path)
        return self.restrict_to_safe_path(data_path)

    @property
    def images_path(self):
        '''All the shared images of the process are located here.'''
        images_path = self.parameters.get(
            'images_path',
            os.path.join(self.process_path, 'images_of_{name}'),
            # os.path.join(self.process_path, 'images'),
        )
        images_path = self.format_with_params('images_path', images_path)
        return self.restrict_to_safe_path(images_path)

    @property
    def analysis_path(self):
        analysis_path = self.parameters.get(
            'analysis_path',
            os.path.join(self.process_path, 'analysis_of_{name}'),
            # os.path.join(self.process_path, 'analysis'),
        )
        analysis_path = self.format_with_params('analysis_path',
                                                analysis_path)
        return self.restrict_to_safe_path(analysis_path)

    @property
    def job_submission_queue(self):
        return self.parameters.get(
            'job_submission_queue',
            SHORT_QUEUE,
        )

    @property
    def job_resubmission_queue(self):
        return self.parameters.get(
            'job_resubmission_queue',
            NORM_QUEUE,
        )

    @property
    def bash_call(self):
        return self.parameters.get(
            'bash_call',
            self.config['tools']['bash']['cmd'],
        )

    @property
    def matlab_call(self):
        return self.parameters.get(
            'matlab_call',
            self.config['tools']['matlab']['cmd'],
        )

    @property
    def python_call(self):
        return self.parameters.get(
            'python_call',
            self.config['tools']['python']['cmd'],
        )

    def get_brainy_lib_path(self, lang):
        return os.path.join(self.config['brainy']['lib_path'],
                            lang.lower())

    @property
    def user_bash_path(self):
        user_path = self.parameters.get('user_bash_path',
                                        '{project_path}/lib/bash')
        return self.format_with_params('user_bash_path', user_path)

    @property
    def user_matlab_path(self):
        user_path = self.parameters.get('user_matlab_path',
                                        '{project_path}/lib/matlab')
        return self.format_with_params('user_matlab_path', user_path)

    @property
    def user_python_path(self):
        user_path = self.parameters.get('user_python_path',
                                        '{project_path}/lib/python')
        return self.format_with_params('user_python_path', user_path)

    def get_user_code_path(self, lang='python', valid_folders=None):
        if valid_folders is None:
            valid_folders = list()
        # Brainy path from deployed code.
        brainy_lib_path = self.get_brainy_lib_path(lang)
        valid_folders.append(brainy_lib_path)
        # User path from config.
        property_name = 'user_%s_path' % lang.lower()
        if hasattr(self, property_name):
            value = getattr(self, property_name)
            user_path = value.split(':')
            for subfolder in user_path:
                if '~/' in subfolder:
                    subfolder = os.path.expanduser(subfolder)
                folder_path = subfolder
                if not os.path.exists(folder_path):
                    logger.warning('The custom code path does not exist: %s' %
                                   folder_path)
                    continue
                valid_folders.append(folder_path)
        # Pack result as colon separated string
        logger.info('Using custom folder path for %s: %s' %
                    (lang, valid_folders))
        return ':'.join(valid_folders)

    def format_with_params(self, param_name, value):
        '''
        Inject updated values into the code. This applies string.format() DSL
        using self.parameters dictionary as arguments.
        '''
        if param_name not in self.compiled_params:
            logger.debug('Compiling (string substitution) param: %s' %
                         param_name)
            process_params = dict()
            for name in self.format_parameters:
                param_var = '{%s}' % name
                if param_var not in value:
                    # Pick only those values that we really need to compile the
                    # parameter - do string substitution.
                    continue
                process_params[name] = getattr(self, name)
            # Compile by substituting every {variable} with its value.
            try:
                compiled_value = value.format(**process_params)
            except KeyError as error:
                message = ('Failed to compile {%s}. Missing value for key'
                           ' %s in expression: "%s".') % \
                          (param_name, str(error), value)
                message += ' Consider doubling the {{ and }} to print { and }.'
                raise BrainyProcessError(message=message,
                                         message_type='warning')
            self.compiled_params[param_name] = compiled_value
        return self.compiled_params[param_name]

    def _get_flag_prefix(self):
        return os.path.join(self.process_path, self.name)

    def get_job_reports(self):
        if self.__reports is None:
            # Find result files only once.
            reports_regex = re.compile(self.job_report_exp)
            self.__reports = [filename for filename
                              in os.listdir(self.reports_path)
                              if reports_regex.search(filename)]
        return self.__reports

    def list_batch_dir(self):
        if self.__batch_listing is None:
            logger.info('Listing batch folder. Please wait.. ')
            self.__batch_listing = list()
            for filename in os.listdir(self.batch_path):
                self.__batch_listing.append(
                    os.path.join(self.batch_path, filename))
            #     sys.stdout.write('.')
            # sys.stdout.write('\n')
            # sys.stdout.flush()
        return self.__batch_listing

    def submit_job(self, shell_command, queue=None, report_file=None,
                   is_resubmitting=False):
        # Differentiate between submission and resubmission parameters.
        if is_resubmitting:
            if not queue:
                queue = self.job_resubmission_queue
        else:
            if not queue:
                queue = self.job_submission_queue
        # Form a unique report file path.
        if not report_file:
            report_file = os.path.join(
                self.reports_path, '%s_%s.job_report' %
                (self.name, datetime.now().strftime('%y%m%d%H%M%S')))
        elif not report_file.startswith('/'):
            report_file = os.path.join(self.reports_path, report_file)
        assert os.path.exists(os.path.dirname(report_file))
        return self.scheduler.submit_job(shell_command, queue, report_file)

    def bake_bash_code(self, bash_code):
        return '''%(bash_call)s << BASH_CODE;
export PATH="%(user_path)s:$PATH"
%(bash_code)s
BASH_CODE''' % {
            'bash_call': self.bash_call,
            'bash_code': format_code(bash_code, lang='bash'),
            'user_path': self.get_user_code_path(lang='bash'),
        }

    def submit_bash_job(self, bash_code, queue=None, report_file=None,
                        is_resubmitting=False):
        script = self.bake_bash_code(bash_code)
        return self.submit_job(script, queue, report_file)

    def bake_matlab_code(self, matlab_code):
        return '''%(matlab_call)s << MATLAB_CODE;
path('%(user_path)s', path());
path(getrecpath('%(user_path)s'), path());

%(matlab_code)s
MATLAB_CODE''' % {
            'matlab_call': self.matlab_call,
            'matlab_code': format_code(matlab_code, lang='matlab'),
            'user_path': self.get_user_code_path(lang='matlab'),
        }

    def submit_matlab_job(self, matlab_code, queue=None, report_file=None,
                          is_resubmitting=False):
        script = self.bake_matlab_code(matlab_code)
        return self.submit_job(script, queue, report_file)

    def bake_python_code(self, python_code):
        user_path = str(self.get_user_code_path(
                        lang='python')).split(':')
        assert type(user_path) == list
        return '''%(python_call)s - << PYTHON_CODE;
import sys
sys.path = %(user_path)s + sys.path
%(python_code)s
PYTHON_CODE''' % {
            'python_call': self.python_call,
            'python_code': format_code(python_code, lang='python'),
            'user_path': user_path,
        }

    def submit_python_job(self, python_code, queue=None, report_file=None,
                          is_resubmitting=False):
        script = self.bake_python_code(python_code)
        return self.submit_job(script, queue, report_file)

    @property
    def job_reports_count(self):
        return len(self.get_job_reports())

    def has_job_reports(self):
        return self.job_reports_count > 0

    def working_jobs_count(self, needle=None):
        if needle is None:
            # TODO: make jobs more specific. Process path is too general.
            needle = os.path.dirname(self.reports_path)
        return self.scheduler.count_working_jobs(needle)

    def put_on(self):
        super(BrainyProcess, self).put_on()
        # Create missing process folder path.
        if not os.path.exists(self.process_path):
            os.makedirs(self.process_path)
        # Create missing job reports folder path.
        if not os.path.exists(self.reports_path):
            os.makedirs(self.reports_path)
        # Set status to
        self.results['step_status'] = 'submitting'

    def want_to_submit(self):
        if self.is_submitted is False and self.is_resubmitted is False \
                and self.is_complete is False:
            return True
        return False

    def no_work_is_happening(self):
        if (self.is_submitted or self.is_resubmitted) \
                and self.is_complete is False \
                and self.working_jobs_count() == 0 \
                and self.has_job_reports() is False \
                and self.has_runlimit is False:
            return True
        return False

    def still_working(self):
        if (self.is_submitted or self.is_resubmitted) \
                and self.is_complete is False:
            return self.has_job_reports() is False \
                and self.working_jobs_count() > 0
        return False

    def has_data(self):
        '''
        Override this method to implement step specific data existence and
        integrity checks.
        '''
        # raise NotImplemented('Missing implementation for has_data() method.')
        return True

    def finished_work_but_has_no_data(self):
        if (self.is_submitted or self.is_resubmitted) \
                and self.is_complete is False \
                and self.has_runlimit() is False:
            # Independent of the fact if we have reports or not, not having any
            # data or jobs running is good enough to attempt resubmission.
            return bool(self.has_data()) is False
        return False

    def report(self, message, warning=''):
        if warning:
            logger.warn(warning)
            BrainyReporter.append_warning(warning)
        logger.info('%(step_name)s: %(message)s %(warning)s' % {
            'step_name': self.step_name,
            'message': message,
            'warning': warning,
        })
        BrainyReporter.append_message(message)

    def has_runlimit(self):
        return os.path.exists(self.get_flag('runlimit'))

    def remove_job_report_file(self, report_filepath):
        '''Remove file from the disk as well as from the cache.'''
        # Clean cache.
        report_filename = os.path.basename(report_filepath)
        if report_filename in self.__reports:
            item_index = self.__reports.index(report_filename)
            del self.__reports[item_index]
        # Remove from disk.
        os.unlink(report_filepath)

    def check_logs_for_errors(self):
        for report_filename in list(self.get_job_reports()):
            # print report_filename
            report_filepath = os.path.join(self.reports_path, report_filename)
            try:
                check_report_file_for_errors(report_filepath)
            except TermRunLimitError as error:
                if self.has_runlimit():
                    message = 'Job %s timed out too many times.' % \
                        report_filename
                    raise BrainyProcessError(
                        message=message.strip(),
                        message_type='warning',
                        output=error.details,
                        job_report=report_filepath,
                    )
                else:
                    BrainyReporter.append_known_error(
                        'Job has exceeded runlimit, resetting job and placing '
                        'timeout flag file')
                    self.set_flag('runlimit')
            except KnownError as error:
                self.reset_resubmitted()
                message = 'Resetting ".(re)submitted" and ".runlimit" flags'\
                    ' and removing job report.'
                # Modifies self.__reports
                self.remove_job_report_file(report_filepath)
                raise BrainyProcessError(
                    message=message,
                    message_type='warning',
                    output=str(error) + error.details,
                    error_type=error.type,
                )
            except UnknownError as error:
                message = 'Unknown error found in job report file %s' % \
                    report_filename
                raise BrainyProcessError(
                    message=message,
                    message_type='warning',
                    output=str(error) + error.details,
                )
        # Finally, no errors were found.

    def resubmit(self):
        self.set_flag('runlimit')

    def check_for_missed_errors(self):
        '''
        "Check for errors" might have failed to detect the error, so we still
        need to make sure that our data was generated correctly.
        '''
        if not self.has_data():
            report_filepath = None
            if self.has_job_reports():
                # Pick latest report
                reports = self.get_job_reports()
                if len(reports) > 0:
                    report_filename = reports[0]
                    report_filepath = os.path.join(self.reports_path,
                                                   report_filename)
            if self.has_runlimit():
                message = 'We reached a limit of retry attempts, but data are'\
                    + ' missing possibly due to some undetected errors. Pleas'\
                    + 'e inspect the corresponding log files.'
                raise BrainyProcessError(message=message,
                                         message_type='warning',
                                         job_report=report_filepath)
            raise BrainyProcessError(message='Data are missing possibly due to'
                                     ' some undetected errors. Please inspect '
                                     'the corresponding log files.',
                                     message_type='warning',
                                     job_report=report_filepath)

    def run(self):
        logger.info('running process <%s> with the following parameters:' %
                    self.name)
        logger.info(pprint.pformat(self.parameters))
        # print self.get_job_reports()
        # print self.working_jobs_count()
        # Skip if ".complete" flag was found.
        if self.is_complete:
            self.results['step_status'] = 'completed'
        # Step is incomplete. Do we want to submit?
        elif self.want_to_submit():
            if self.working_jobs_count() > 0:
                logger.warn('Some jobs are still running, but we are '
                            'submitting everything a new!')
            self.submit()
        # Submitted but no work has started yet?
        elif self.no_work_is_happening():
            self.report('resetting', warning='No work is happening')
            self.reset_submitted()
        # Waiting.
        elif self.still_working():
            self.report('waiting')
        # Resubmit if no data output but job results found.
        elif self.finished_work_but_has_no_data():
            # Hint: check if has_data() works correctly!
            self.resubmit()
        # Check for known errors if both data output and job results are
        # present.
        else:
            logger.info('Checking logs for errors. Setting process state to '
                        'completed if none found.')
            self.check_logs_for_errors()
            self.check_for_missed_errors()
            # At this point we have detected no errors and the data test
            # has passed -> mark process as 'completed'.
            self.results['step_status'] = 'completed'
            self.set_flag('complete')

    def reduce(self):
        if self.results['step_status'] == 'completed':
            logger.info('Step {%s} was completed.' % self.step_name)
            BrainyReporter.append_message('complete', status='complete')



'''
brainy.project

Manage projects.

@author: Yauhen Yakimovich <yauhen.yakimovich@uzh.ch>,
         Pelkmans Lab <https://www.pelkmanslab.org>

@license: The MIT License (MIT)

Copyright (c) 2014 Pelkmans Lab
'''

import os
import shutil
import logging
from brainy.version import brainy_version
from brainy.utils import merge_dicts
from brainy.config import (write_project_config, load_project_config,
                           load_user_config, project_has_config)
from brainy.workflows import bootstrap_workflow
from brainy.scheduler import BrainyScheduler
from brainy.pipes.manager import PipesManager
from brainy.errors import BrainyProjectError
from brainy.project.report import report_data
logger = logging.getLogger(__name__)


class BrainyProject(object):

    def __init__(self, name, path):
        self.name = name
        if os.path.basename(path) != name:
            # Append name to full path of the project folder if it is not
            # ending with such.
            self.path = os.path.join(path, name)
        else:
            self.path = path
        self.config = None
        self.scheduler = None

    @property
    def report_folder_path(self):
        return os.path.join(self.path, 'reports')

    @property
    def report_prefix_path(self):
        return os.path.join(self.report_folder_path, self.name)

    def seed_report_data(self):
        '''Put meta data about this project into report'''
        logger.info('Seeding report data.')
        report_data['project'] = {
            'brainy_version': brainy_version,
            'name': self.name,
            'pipes': [],
        }

    def init(self, inherit_config={}, override_config={}):
        # Make project dir.
        if not os.path.exists(self.path):
            raise BrainyProjectError('Project folder does not exists: %s' %
                                     self.path)
        # Put basic YAML config during project creation. This is required for
        # project to be valid. We will also check version compatibility.
        write_project_config(self.path, inherit_config=inherit_config,
                             override_config=override_config)

    def create(self, from_workflow='canonical', inherit_config={},
               override_config={}):
        # Make project dir.
        if not os.path.exists(self.path):
            logger.info('Creating new project folder: %s' % self.path)
            os.mkdir(self.path)
        else:
            raise BrainyProjectError('Project folder already exists: %s' %
                                     self.path)
        self.init(inherit_config=inherit_config,
                  override_config=override_config)
        # Bootstrap project with a standard iBRAIN workflow.
        bootstrap_workflow(self.path, workflow_name=from_workflow)

    def is_a_valid_project_folder(self):
        return project_has_config(self.path)

    def load_config(self):
        self.config = load_user_config()
        project_config = load_project_config(self.path)
        logger.info('Overwriting global configuration with project specific.')
        self.config = merge_dicts(self.config, project_config)

    def clean(self, manager_cls=PipesManager):
        '''Clean the project house.'''
        logger.info('<brainy rolls over the project to clean the house>')
        logger.info('Loading configuration')
        self.load_config()
        self.pipes = manager_cls(self)
        logger.info('Removing all the output data of every pipeline.')
        self.pipes.run('clean_pipelines_output')
        if os.path.exists(self.report_folder_path):
            logger.info('Removing all the reports in: %s' %
                        self.report_folder_path)
            shutil.rmtree(self.report_folder_path)
        logger.info('<Done>')

    def run(self, manager_cls=PipesManager, command='process_pipelines'):
        '''
        Load project configuration. Discover and process pipelines using the
        specified scheduler.
        '''
        logger.info('<brainy rolls over the project to {%s}>' % command)
        logger.info('Loading configuration')
        self.load_config()
        self.seed_report_data()
        logger.info('Initializing "%s" as a scheduling engine.' %
                    self.config['scheduling']['engine'])
        self.scheduler = BrainyScheduler.build_scheduler(
            self.config['scheduling']['engine'])
        self.pipes = manager_cls(self)
        self.pipes.run(command)
        logger.info('<Done>')



import logging
logger = logging.getLogger(__name__)


# TODO: Queuing settings have to be configurable
SHORT_QUEUE = '1:00'
NORM_QUEUE = '8:00'
LONG_QUEUE = '36:00'


# Job states
PENDING_STATE = 1
RUNNING_STATE = 2
DONE_STATE = 3
JOB_STATES = [PENDING_STATE, RUNNING_STATE, DONE_STATE]


class BrainyScheduler(object):
    '''
    Base scheduling engine class. In general, scheduler is responsible for
    submitting, monitoring, killing and performing other job managing
    operations.
    '''

    @staticmethod
    def build_scheduler(name):
        '''Factory for scheduling engines.'''
        name = name.lower()
        if name == 'lsf':
            from brainy.scheduler.lsf import Lsf
            return Lsf()
        elif name == 'shellcmd':
            from brainy.scheduler.shellcmd import ShellCommand
            return ShellCommand()
        raise Exception('Unknown scheduler type: %s' % name)




import os
import re
from fnmatch import fnmatch, translate as fntranslate
from os.path import basename
from brainy.process import BrainyProcess, BrainyProcessError
from brainy.pipes import BrainyPipe
from brainy.pipes.Tools import get_timestamp_str


class Pipe(BrainyPipe):
    '''
    Jterator pipe includes steps like:
     - Create job list and pre-cluster (run the first job of the pipeline)
     - RunInParallel
     - MergeResults
    '''


class PreCluster(BrainyProcess):
    '''Create job list and Jterate over the first job'''


class RunInParallel(BrainyProcess):
    '''Jterate over the rest of jobs in parallel'''


class MergeResults(BrainyProcess):
    '''Merge/reduce output results by job'''



'''
brainy.config

Save and load configuration files.

@author: Yauhen Yakimovich <yauhen.yakimovich@uzh.ch>,
         Pelkmans Lab  <https://www.pelkmanslab.org>

@license: The MIT License (MIT)

Copyright (c) 2014 Pelkmans Lab
'''
import os
import yaml
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
from getpass import getuser
import logging
logger = logging.getLogger(__name__)
from brainy.version import brainy_version
from brainy.utils import merge_dicts


# User-wide (global) configuration
BRAINY_USER_CONFIG_TPL = '''
# brainy user config
brainy:
    version: '%(brainy_version)s'
    user: '%(brainy_user)s'
    lib_path: '%(brainy_lib_path)s'
    pipe_extension: '.br'
    admin_email: 'root@localhost'

# Which scheduling API to use by default?
scheduling:
    # Possible choices are: {'shellcmd', 'lsf', 'slurm'}
    engine: 'shellcmd'

# Preliminary tools and programming languages
tools:
    python:
        cmd: '/usr/bin/env python2.7'
    matlab:
        cmd: '/usr/bin/env matlab -singleCompThread -nodisplay -nojvm'
    ruby:
        cmd: '/usr/bin/env ruby'
    node:
        cmd: '/usr/bin/env node'
    bash:
        cmd: '/bin/bash'


# Integrated application
apps:
    CellProfiler2:
        path: '%(cellprofiler2_path)s'

# Default project parameters
project_parameters:
    job_submission_queue: '8:00'
    job_resubmission_queue: '36:00'

''' % {
    'brainy_version': brainy_version,
    'brainy_user': getuser(),
    'cellprofiler2_path': os.path.expanduser(
        '~/iBRAINFramework/tools/CellProfiler2'),
    'brainy_lib_path': os.path.join(
        os.path.dirname(__file__),
        'lib',
    ),
}
BRAINY_USER_CONFIG_PATH = os.path.expanduser('~/.brainy/config')


# Project specific configuration
BRAINY_PROJECT_CONFIG_TPL = '''
# brainy project config
brainy:
    version: '%(brainy_version)s'

# Parameters below will affect the whole project. In particular, that defines
# how jobs in every step of each pipe will be submitted, which folder names
# and locations are used, etc.

# Uncomment these section to use parallel submission engines once your project
# is deployed onto a cluster.
#
# scheduling
#     engine: 'lsf'

project_parameters:
    # job_submission_queue: '8:00'
    # job_resubmission_queue: '36:00'
    batch_path: 'data_of_{name}'
    tiff_path: 'images_of_{name}'

''' % {
    'brainy_version': brainy_version,
}
BRAINY_PROJECT_CONFIG_NAME = '.brainy'


def write_config(config_path, value):
    logger.info('Writing config: %s' % config_path)
    if type(value) == dict:
            value = yaml.dump(value, default_flow_style=True)
    with open(config_path, 'w+') as stream:
        stream.write(value)


def load_config(config_path):
    logger.info('Loading config: %s' % config_path)
    with open(config_path) as stream:
        return yaml.load(stream, Loader=Loader)


def write_user_config(user_config_path=BRAINY_USER_CONFIG_PATH):
    '''Write global config into '.brainy' folder inside user's home.'''
    user_brainy_path = os.path.dirname(user_config_path)
    if not os.path.exists(user_brainy_path):
        logger.info('Creating a missing user brainy folder: %s' %
                    user_brainy_path)
        os.makedirs(user_brainy_path)
    if os.path.exists(user_config_path):
        logger.warn('Abort. Configuration file already exists: %s' %
                    user_config_path)
        return
    write_config(user_config_path, BRAINY_USER_CONFIG_TPL)


def load_user_config():
    return load_config(BRAINY_USER_CONFIG_PATH)


def write_project_config(project_path, config_name=BRAINY_PROJECT_CONFIG_NAME,
                         inherit_config={}, override_config={}):
    config_path = os.path.join(project_path, config_name)
    if inherit_config or override_config:
        value = yaml.load(BRAINY_PROJECT_CONFIG_TPL)
        if inherit_config:
            value = merge_dicts(inherit_config, value)
        if override_config:
            value = merge_dicts(value, override_config)
    else:
        value = BRAINY_PROJECT_CONFIG_TPL
    write_config(
        config_path=config_path,
        value=value,
    )


def load_project_config(project_path, config_name=BRAINY_PROJECT_CONFIG_NAME):
    config_path = os.path.join(project_path, config_name)
    return load_config(config_path)


def project_has_config(project_path, config_name=BRAINY_PROJECT_CONFIG_NAME):
    config_path = os.path.join(project_path, config_name)
    return os.path.exists(config_path)



from brainy.errors import BrainyProcessError


def require_keys_in_description(*description_keys):
    '''
    Apply this decorator to process class if you want to require process
    descriptor to contain particular keys. Notice that this will inject
    respective *property* into the process description.
    '''

    def add_required_property(cls, property_name):
        assert not hasattr(cls, property_name)

        def get_required_description_key(self):
            '''Require process descriptor to contain: %s''' % property_name
            try:
                return self.description[property_name]
            except KeyError:
                raise BrainyProcessError(
                    'Missing "%s" key in YAML descriptor of the process.' %
                    property_name
                )

        setattr(cls, property_name,
                property(fget=get_required_description_key))

    def class_builder(original_class):
        for key in description_keys:
            add_required_property(original_class, key)
        return original_class

    return class_builder


def require_key_in_description(method):
    '''
    Same as previous but applicable per property, not class. Make sure
    this decorator is the first to be applied.
    '''
    param_name = method.__name__
    if hasattr(method, 'param_name'):
        param_name = method.param_name

    def get_required_description_key(self):
        if param_name not in self.description:
            raise BrainyProcessError(
                'Missing "%s" key in YAML descriptor of the process.' %
                param_name
            )
        result = method(self)
        return self.description[param_name] if result is None else result

    get_required_description_key.param_name = param_name
    return get_required_description_key


def format_with_params(method):
    '''
    Apply this decorator to properties of BrainyProcess() class instances and
    its descendants if you wish to substitute string values of all the
    parameters specified as {variable} using format_with_params().
    '''
    param_name = method.__name__
    if hasattr(method, 'param_name'):
        param_name = method.param_name

    def formatting_with_params(self):
        return self.format_with_params(param_name,
                                       method(self))

    formatting_with_params.param_name = param_name
    return formatting_with_params



import os
import re
import textwrap
from fnmatch import translate as fntranslate
import brainy
from brainy.errors import BrainyProcessError
from brainy.process.code import PythonCodeProcess
from brainy.pipes.base import BrainyPipe
from brainy.config import get_config


brainy_config = get_config()


def get_cp2_call():
    return '%s %s' % (
        brainy_config['python_cmd'],
        os.path.join(os.path.expanduser(brainy_config['cellprofiler2_path']),
                     'CellProfiler.py'),
    )


def create_imagelists_for_batching(tiff_path, batch_path,
                                   image_list_settings_filename):
    ''' Create CSV input image lists'''
    from brainy.apps.cellprofiler import CellProfilerImages

    images_path = tiff_path
    output_path = os.path.join(batch_path, 'IMAGE_LISTS')
    if os.path.exists(batch_path)\
            and not os.path.exists(output_path):
        os.makedirs(output_path)
    image_list_settings_filename = image_list_settings_filename

    cpimages = CellProfilerImages()
    if os.path.exists(image_list_settings_filename):
        # Load custom JSON settings file
        cpimages.parse_settings(image_list_settings_filename)
    cpimages.split_images(images_path, output_path)
    num_of_image_sets = cpimages.set_num

    if num_of_image_sets == 0:
        print 'Error, splitting of images failed. No image set were '\
              'generated.'
        exit(1)

    return cpimages


def run_cp2_pipeline_batch(tiff_path, batch_path, cp_pipeline_file,
                           csv_filepath):
    '''
    Execute CellProfiller2 in process in a command line. Pass arguments
    sufficient to process a single batch of images.
    '''
    command_lines = textwrap.wrap('''
    %(cp2_call)s -b -c -i %(tiff_path)s -o %(batch_path)s \
        --do-not-build --do-not-fetch --pipeline=%(cp_pipeline_file)s \
        --data-file=%(csv_filepath)s -L INFO
    ''' % {
        'cp2_call': get_cp2_call(),
        'tiff_path': tiff_path,
        'batch_path': batch_path,
        'cp_pipeline_file': cp_pipeline_file,
        'csv_filepath': csv_filepath,
    }, width=210, break_on_hyphens=False, break_long_words=False)
    command = ' \\\n'.join(command_lines)
    print command
    return brainy.invoke(command)


class Pipe(BrainyPipe):
    '''
    CellProfiller 2.1 pipe includes steps like:
     - CreateJobBatches
     - SubmitJobs
     - MergeJobData
    '''


class CreateJobBatches(PythonCodeProcess):
    '''
    Run CellProfiller2 project (pipeline) with CreateBatchFiles module.
    Expecting to create BATCH/Batch_data.h5 file.
    '''

    @property
    def cp_pipeline_fnpattern(self):
        return self.description.get(
            'filename',
            'run_*.cppipe',
        )

    def get_cp_pipeline_path(self):
        filename_regex_obj = re.compile(fntranslate(
                                        self.cp_pipeline_fnpattern))
        cp_pipeline_files = [
            filename for filename in os.listdir(self.process_path)
            if filename_regex_obj.match(filename)
        ]
        if len(cp_pipeline_files) > 1:
            raise BrainyProcessError('More than one CP pipeline settings file'
                                     ' found matching: %s in %s' %
                                     (self.cp_pipeline_fnpattern,
                                      self.process_path))
        elif len(cp_pipeline_files) == 0:
            raise BrainyProcessError('No CP pipeline settings file'
                                     ' found matching: %s in %s' %
                                     (self.cp_pipeline_fnpattern,
                                      self.process_path))
        return os.path.join(self.process_path, cp_pipeline_files[0])

    def put_on(self):
        super(CreateJobBatches, self).put_on()
        # Make sure those pathnames exists. Create if missing.
        if not os.path.exists(self.batch_path):
            os.makedirs(self.batch_path)
        if not os.path.exists(self.postanalysis_path):
            os.makedirs(self.postanalysis_path)

    def get_python_code(self):
        '''
        Note that the interpreter call is included by
        self.submit_python_code()
        '''
        code = '''
        from brainy.pipes.CellProfiler2 import (
            create_imagelists_for_batching,
            run_cp2_pipeline_batch)

        cpimages = create_imagelists_for_batching(
            '%(tiff_path)s',
            '%(batch_path)s',
            '%(image_list_settings_filename)s',
        )

        # Test pipeline by running first batch of images
        output = run_cp2_pipeline_batch(
            '%(tiff_path)s',
            '%(batch_path)s',
            '%(cp_pipeline_file)s',
            # Pick first set of images for this test run.
            csv_filepath=cpimages.saved_csv_files[0],
        )
        # TODO: check output for any errors or exceptions raised.
        # But do it rather in has_data().
        print output
        ''' % {
            'tiff_path': self.tiff_path,
            'batch_path': self.batch_path,
            'image_list_settings_filename': self.description.get(
                'image_list_settings_filename',
                '',
            ),
            'cp_pipeline_file': self.get_cp_pipeline_path(),
        }
        return code




import os
import logging
logger = logging.getLogger(__name__)
from sh import touch
from brainy.pipes import PipesModule


class CompatiblePipesModule(PipesModule):
    '''Brainy pipes module with backwards compatibility with iBRAIN.'''

    @property
    def stage_complete_flagpath(self):
        return os.path.join(self.env['plate_path'],
                            'iBRAIN_Stage_1.completed')

    def complete_stage_one(self):
        '''
        Try to complete stage_one phase of iBRAIN by creating a respective
        flag, so that the rest of iBRAIN logic can kick-in the processing.
        '''
        if os.path.exists(self.stage_complete_flagpath):
            logger.info('Found a flag. Stage one is completed.')
            return
        all_pipelines_are_complete = True
        for pipeline in self.pipelines:
            for process in pipeline.chain:
                if not process.is_complete:
                    all_pipelines_are_complete = False
                    break
            if not all_pipelines_are_complete:
                break
        # All pipelines are complete.
        if all_pipelines_are_complete:
            logger.info('All pipelines of stage one are complete. '
                        'Creating a flag.')
            touch(self.stage_complete_flagpath)
        else:
            logger.info('Not all pipelines of stage one are complete.')



import re


MAX_ERROR_MSG_SIZE = 1000


class UnknownError(Exception):
    '''Unknown error found in iBRAIN job report'''


class KnownError(Exception):
    '''Known error found in iBRAIN job report'''

    def __init__(self, message, details=None, error_type=None):
        super(KnownError, self).__init__(message)
        self.details = details
        self.type = error_type


class TermRunLimitError(KnownError):
    '''Job exceeded time limit, resetting job and placing timeout flag file.'''


class OutOfMemoryError(KnownError):
    '''
    Job exceeded memory limit, resetting job and placing timeout flag file.
    '''


#UNKNOWN_ERROR = re.compile('ERROR', re.IGNORECASE)
UNKNOWN_ERROR = re.compile('Exited with exit code', re.IGNORECASE)


KNOWN_ERRORS = {
    # type -> (needle, explanation)
    'out_of_time': {
        'token': 'TERM_RUNLIMIT',
        'cause': 'Timed out too many times',
    },
    'out_of_memory': {
        #'token': '"Out of memory. Type HELP MEMORY for your options."',
        'token': 'Out of memory.',
        'cause': 'Job exceeded memory limit',
    },
    'job_terminated': {
        'token': 'TERM_OWNER',
        'cause': 'Owner terminated job',
    },
    'ghostscript_call_failed': {
        'token': 'Problem calling GhostScript',
        'cause': 'Problem calling GhostScript',
    },
    'matlab_no_license_1': {
        'token': 'License Manager Error',
        'cause': 'Matlab License Manager Error',
    },
    'matlab_no_license_2': {
        'token': 'License Manager Error -15',
        'cause': 'Matlab License Manager Error',
    },
    'matlab_function_not_found': {
        'token': 'Undefined function or method',
        'cause': 'MATLAB: Undefined function or method',
    },
    'matlab_command_not_found': {
        'token': 'matlab: command not found',
        'cause': 'Environment error: Matlab command not found',
    },
    'matlab_segmentation_violation': {
        'token': 'Segmentation violation',
        'cause': 'Abnormal termination: Segmentation violation',
    },
    'hdf5_io_error': {
        'token': 'Error detected in HDF5',
        'cause': 'Error loading HDF5 file',
    },
}


def grab_details(text, token):
    start_pos = text.find(token)
    details_span = start_pos + MAX_ERROR_MSG_SIZE
    return text[start_pos:details_span]


def check_for_known_error(text):
    for error_type in KNOWN_ERRORS:
        error = KNOWN_ERRORS[error_type]
        if error['token'] in text:
            if error_type == 'out_of_time':
                raise TermRunLimitError(
                    '[KNOWN] %s' % error['cause'],
                    details=grab_details(text, error['token']),
                    error_type=error_type)
            elif error_type == 'out_of_memory':
                raise OutOfMemoryError(
                    '[KNOWN] %s' % error['cause'],
                    details=grab_details(text, error['token']),
                    error_type=error_type)
            else:
                raise KnownError(
                    '[KNOWN] %s' % error['cause'],
                    details=grab_details(text, error['token']),
                    error_type=error_type)
    # Empty error message <-> no error.
    return ''


def check_report_file_for_errors(report_filepath):
    report_text = open(report_filepath).read()
    try:
        check_for_known_error(report_text)
    except KnownError as error:
        raise
    # Has unknown error?
    match = UNKNOWN_ERROR.search(report_text)
    if not match:
        return
    error = UnknownError('[UNKNOWN] Unknown error found')
    error.details = grab_details(report_text, match.group(0))
    raise error


class BrainyProjectError(Exception):
    '''Thrown by brainy project if the logic goes wrong.'''


class BrainyProcessError(Exception):
    '''Logical error that happened while executing brainy pipe process'''

    def __init__(self, message, message_type='error', output='',
                 job_report=None, **kwds):
        super(BrainyProcessError, self).__init__(message)
        self.extra = {}
        self.extra['message_type'] = message_type  # info, warning, error
        self.extra['output'] = output
        self.extra['job_report'] = job_report
        self.extra.update(kwds)


class BrainyPipeFailure(Exception):
    '''Thrown when pipeline execution has to be interrupted.'''


class ProccessEndedIncomplete(BrainyPipeFailure):
    '''One of the pipe's processes failed to complete successfully.'''

    def __init__(self, process):
        self.process = process

    def __str__(self):
        return self.__doc__



'''
iBRAINPipes is an integration of pipette processes into iBRAIN modules.
'''
from __future__ import with_statement
import os
import pipette
import logging
import pprint
from brainy.errors import (BrainyProcessError, ProccessEndedIncomplete,
                           BrainyPipeFailure)
from brainy.project.report import BrainyReporter
logger = logging.getLogger(__name__)


PROCESS_NAMESPACE = 'brainy.pipes'


class BrainyPipe(pipette.Pipe):

    def __init__(self, pipes_manager, definition=None):
        super(BrainyPipe, self).__init__(PROCESS_NAMESPACE, definition)
        self.pipes_manager = pipes_manager
        self.has_failed = False
        self.previous_process_params = None

    @property
    def pipe_extension(self):
        return self.pipes_manager.pipe_extension

    @property
    def output_path(self):
        return os.path.join(self.pipes_manager.project_path, self.name)

    def instantiate_process(self, process_description,
                            default_type=None):
        process = super(BrainyPipe, self).instantiate_process(
            process_description, default_type)
        process.name_prefix = self.name
        return process

    def get_step_name(self, process_name):
        return '%s-%s' % (self.name, process_name)

    def get_previous_parameters(self):
        if self.previous_process_params is None:
            return
        if not self.previous_process_params['previous_process_params'] is None:
            # Avoid chaining the back up to the first process. Such linking
            # can motivate a very bad programming practices. Only one step
            # before is allowed to memorize. Everything else is just to
            # complicated. So we unlink previous of previous here.
            self.previous_process_params['previous_process_params'] = None
        return self.previous_process_params

    def get_process_path(self):
        return os.path.join(
            self.pipes_manager._get_flag_prefix(),
            self.name,
        )

    def get_process_parameters(self):
        parameters = dict()
        parameters['pipes_manager'] = self.pipes_manager
        parameters['process_path'] = self.get_process_path()
        # Update them with project-wide parameters
        project = self.pipes_manager.project
        logger.info('Applying project-wide parameters.')
        logger.debug(pprint.pformat(project.config['project_parameters']))
        parameters.update(project.config['project_parameters'])
        return parameters

    def execute_process(self, process, parameters):
        '''
        Execute process as a step in brainy pipeline. Add verbosity, e.g.
        report status using brainy project report scheme.
        '''
        parameters = self.get_process_parameters()
        # Set step name.
        step_name = self.get_step_name(process.name)
        logger.info('Executing step {%s}' % step_name)
        parameters['step_name'] = self.get_step_name(process.name)
        # Some modules are allowed to have limited dependency on previous
        # steps, but this is restricted. Also check unlinking in
        # get_previous_parameters().
        parameters['previous_process_params'] = self.get_previous_parameters()
        self.previous_process_params = parameters

        BrainyReporter.append_report_process(process.name)

        try:
            super(BrainyPipe, self).execute_process(process, parameters)

        except BrainyProcessError as error:
            # See brainy.errors
            error_is_fatal = False
            if 'message_type' in error.extra:
                if error.extra['message_type'] == 'error':
                    error_is_fatal = True
                del error.extra['message_type']
            if 'error_type' in error.extra:
                BrainyReporter.append_known_error(
                    message=str(error),
                    **error.extra)
            else:
                BrainyReporter.append_unknown_error(
                    message=str(error),
                    **error.extra)
            # Finally, interrupt execution if we error is fatal (default).
            if error_is_fatal:
                logger.exception(error)
                if 'job_report' in error.extra:
                    logger.error('Inspect error details in job report: %s' %
                                 error.extra['job_report'])
                raise BrainyPipeFailure('Execution failed')

        finally:

            if not process.is_complete:
                raise ProccessEndedIncomplete(process=process)



#!/usr/bin/env python2.7
'''
iBRAINTool is helper command line utility class for simplifying handling of
iBRAIN project flags, files, etc. This includes basic operations like refresh
the project modification date or restarting the modules.

Requires PyPlato: https://github.com/ewiger/plato
sudo easy_install PyPlato

This code is distributed under the GNU General Public License.
See the accompanying file LICENSE for details.

Authors:
    Yauhen Yakimovich <eugeny.yakimovitch@gmail.com>,

Copyright (c) 2013 Pelkmans' Lab http://www.pelkmanslab.org/

    Installation
    ============

    Linux (Ubuntu)
    --------------

Use *easy_install* tool in command line:

    sudo easy_install PyPlato

or if you don't have a super-user access

    pip install --user PyPlato


'''
from __future__ import print_function
import os
import re
import sys
import argparse
import shutil
import logging
from plato.shell.findutils import (Match, find_files)
from fnmatch import fnmatch


__version__ = '0.1.1'

DRY_RUN = False
#DRY_RUN = True

VERBOSITY_LEVELS = {
    0: logging.NOTSET,
    1: logging.WARN,
    2: logging.INFO,
    3: logging.DEBUG,
}

# Note that flags support fnmatch format
RESTART_MODULES = {
    'tiff2png': {
        'project_files': [
            'ConvertAllTiff2Png.submitted',
        ],
        'batch_files': [
            'batch_png_convert_*',
            'ConvertAllTiff2Png_batch_png_convert_*_*.results',
        ],
    },
    'illcor': {
        'project_files': [
            'IllumCorrection_batch_illcor_*.submitted',
            'IllumCorrectionSettingsFileCreation.submitted',
        ],
        'batch_files': [
            'IllumCorrection_batch_illcor*.results',
            'Measurements_batch_illcor_*.mat',
            'batch_illcor_*.mat',
            'illuminationcorrection.complete',
        ],
    },
    'svm': {
        'project_files': [
            'SVMClassification_SVM_*.submitted',
            'SVMClassification_SVM_*.resubmitted',
            'SVMClassification_SVM_*.runlimit',
        ],
        'batch_files': [
            'SVMClassification_SVM_*.results',
            # Also kills results!
            'Measurements_SVM_*.mat',
        ],
    },
    'precluster': {
        'project_files': [
            'PreCluster.submitted',
            'PreCluster.resubmitted',
            'PreCluster.runlimit',
        ],
        'batch_files': [
            'PreCluster_*.results',
        ],
    },
    # Reset batch jobs step
    'batchjobs': {
        'project_files': [
            'SubmitBatchJobs.resubmitted',
            'SubmitBatchJobs.submitted',
        ],
        'batch_files': [
            'Batch_*_OUT.mat',
            'Batch_*_to_*_Measurements_*.mat',
            re.compile('Batch_\d+_to_\d+\.mat'),
            re.compile('Batch_\d+_to_\d+\.txt'),
            'Batch_*_to_*_*_*.results',
        ],
    },
    'celltype-overview': {
        'project_files': [
            'CreateCellTypeOverview.submitted',
            'CreateCellTypeOverview.resubmitted',
        ],
        'batch_files': [
            'CreateCellTypeOverview_*.results',
        ],
    },
    'fusion': {
        'project_files': [
        ],
        'batch_files': [
            'DataFusion_Measurements_*.results',
        ],
    },
    'pop-context': {
        'project_files': [
        ],
        'batch_files': [
            'getLocalCellDensityPerWell_auto_*.results',
        ],
    },
    'stich-seg': {
        'project_files': [
        ],
        'batch_files': [
            'StitchSegmentationPerWell_*.results',
        ],
    },
    'plate-overview': {
        'project_files': [
            'CreatePlateOverview.submitted',
        ],
        'batch_files': [
            'CreatePlateOverview_*.results',
        ],
    },
}


# Resume work on those actions. This is similar to restart but does not kills
# previous results in .MAT files.
RESUBMIT_MODULES = {
    'batchjobs': {
        'project_files': [
            'CreateCellTypeOverview.submitted',
            'CreateCellTypeOverview.resubmitted',
        ],
    },
    'svm': {
        'project_files': [
            'SVMClassification_SVM_*.submitted',
            'SVMClassification_SVM_*.resubmitted',
            'SVMClassification_SVM_*.runlimit',
        ],
        'batch_files': [
            'SVMClassification_SVM_*.results',
        ],
    },
}

RESUBMIT_CHOICES = list(RESUBMIT_MODULES.keys())

RESTART_CHOICES = list(RESTART_MODULES.keys()) + ['project']


# Kill those files to get a semi-"vanilla" state of the project
RESET_LIST = {
    'project': {
        'project_files': [
            '*.submitted',
            '*.resubmitted',
            'iBRAIN_Stage_1.completed',
            'FuseBasicData_*.results',
            'BASICDATA*',
            'ADVANCEDDATA*.mat',
        ],
        'project_folders': [
            'BATCH',
            'JPG',
            'POSTANALYSIS',
            'SEGMENTATION*',
        ],
    }
}


def resolve_filenames(patterns, parent_path, recursive=False, filetype='f'):
    '''
    Files can be a list of filenames, fnmatch patterns or regexps

    >>> len(resolve_filenames(['*'], '/tmp/a')) > 0
    True
    '''
    if not os.path.exists(parent_path):
        raise Exception('Failed to resolve: ' + parent_path)
    resolved = list()
    for pattern in patterns:
        # We look only into files for now.
        match = Match(filetype=filetype, name=pattern)
        resolved.extend([fname for fname
                        in find_files(parent_path, match,
                        recursive=recursive)])
    return resolved


class iBRAINToolError(Exception):
    '''Logical error thrown by iBRAINTool'''


class iBRAINTool(object):
    '''
    iBRAINTool is helper utility class for iBRAIN project.

    Any existing folder containing a TIFF is considered to be a valid
    project folder

    >>> project_dir = '/tmp/Project'
    ... os.makedirs(os.path.join(project_dir, '/TIFF'))
    ... tool = iBRAINTool(project_dir)
    '''
    def __init__(self, project_dir, is_multiplate, verbose=logging.INFO):
        logging.basicConfig(format='%(message)s', level=verbose)
        self.logger = logging.getLogger('iBRAINTool')
        if not self.is_project_dir(project_dir, is_multiplate):
            raise iBRAINToolError('Path is not a valid project folder: %s' %
                                  project_dir)
        self.logger.debug('iBRAINTool v%s', __version__)
        self.project_dir = project_dir
        self.is_multiplate = is_multiplate
        self.is_verbose = verbose
        self.__batch_listing = None

    def is_project_dir(self, path, is_multiplate):
        if not os.path.exists(path):
            raise IOError('Path does not exists: %s' %
                          path)
        if is_multiplate:
            return len(self.find_plate_names(path)) > 0
        else:
            return self.is_plate_dir(path)

    def find_plate_names(self, path):
        return [name for name in os.listdir(path)
                if self.is_plate_dir(os.path.join(path, name))]

    def is_plate_dir(self, path):
        tiff_path = os.path.join(path, 'TIFF')
        nikon_path = os.path.join(path, 'NIKON')
        return any(os.path.exists(plate_path) and os.path.isdir(plate_path)
                   for plate_path in (tiff_path, nikon_path))

    @property
    def batch_dir(self):
        return os.path.join(self.project_dir, 'BATCH')

    def kill_files(self, files):
        for filepath in files:
            if not DRY_RUN:
                print('Killing ' + filepath)
                os.unlink(filepath)
            else:
                print('Will kill ' + filepath)

    def kill_folders(self, folders):
        for folderpath in folders:
            if not DRY_RUN:
                print('Recursively killing ' + folderpath)
                shutil.rmtree(folderpath)
            else:
                print('Will recursively kill ' + folderpath)

    def restart_module(self, modulename):
        self.logger.info('Restarting module \'%s\' in project: %s' %
                         (modulename, self.project_dir))
        if not modulename in RESTART_MODULES:
            raise iBRAINToolError('Unknown module')
        module = RESTART_MODULES[modulename]
        project_files = resolve_filenames(
            module.get('project_files', []), self.project_dir)
        self.kill_files(project_files)
        batch_files = resolve_filenames(
            module.get('batch_files', []), self.batch_dir)
        self.kill_files(batch_files)
        self.restart_project()

    def resubmit_module(self, modulename):
        self.logger.info('Resubmitting module \'%s\' in project: %s' %
                         (modulename, self.project_dir))
        if not modulename in RESUBMIT_MODULES:
            raise iBRAINToolError('Unknown module')
        module = RESUBMIT_MODULES[modulename]
        project_files = resolve_filenames(
            module.get('project_files', []), self.project_dir)
        self.kill_files(project_files)
        batch_files = resolve_filenames(
            module.get('batch_files', []), self.batch_dir)
        self.kill_files(batch_files)
        self.restart_project()

    def restart_project(self):
        if DRY_RUN:
            self.logger.info('Will restart project: %s' % self.project_dir)
            return
        self.logger.info('Restarting project: %s' % self.project_dir)
        # Set modification time to current time.
        os.utime(self.project_dir, None)

    def reset_project(self, confirm=False):
        if not confirm:
            raise iBRAINToolError('Reset action requires confirmation, e.g. '
                                  'you can use --force action')
        self.logger.info('Reseting project: %s' % self.project_dir)
        # Kill folders and files.
        for key in RESET_LIST:
            kill_files = RESET_LIST[key].get('project_files', [])
            self.kill_files(resolve_filenames(kill_files, self.project_dir))
            kill_folders = RESET_LIST[key].get('project_folders', [])
            self.kill_folders(resolve_filenames(
                kill_folders, self.project_dir, filetype='d',
            ))
        self.restart_project()

    def list_batch_dir(self):
        if self.__batch_listing is None:
            print('Listing batch folder. Please wait: ')
            self.__batch_listing = list()
            for filename in os.listdir(self.batch_dir):
                self.__batch_listing.append(
                    os.path.join(self.batch_dir, filename))
                sys.stdout.write('.')
            sys.stdout.write('\n')
            sys.stdout.flush()
        return self.__batch_listing

    def resubmit(self, modulename):
        if modulename == 'batchjobs':
            from os.path import basename
            expr = re.compile('Batch_(\d+_to_\d+)')
            parse_batch = lambda filename: expr.search(basename(filename))\
                                           .group(1)
            output_batches = [parse_batch(filename) for filename
                              in self.list_batch_dir()
                              if fnmatch(basename(filename),
                                         'Batch_*_OUT.mat')]
            result_batches = dict(((parse_batch(filename), filename)
                                  for filename in self.list_batch_dir()
                                  if fnmatch(basename(filename),
                                             'Batch_*.results')))
            print('Found %d job results logs vs. %d .MAT output files' %
                  (len(result_batches), len(output_batches)))
            # Check if every result has a corresponding batch output.
            # If not, consider it to be an error.
            error_results = list()
            for batch_span in result_batches:
                if not batch_span in output_batches:
                    print('Resubmitting incomplete batch: %s' % batch_span)
                    error_results.append(result_batches[batch_span])
            if error_results:
                # Kill found errors
                self.kill_files(error_results)
                # Kill submission flags
                submit_flags = resolve_filenames([
                    'SubmitBatchJobs.submitted',
                    'SubmitBatchJobs.resubmitted',
                    'SubmitBatchJobs.runlimit',
                ], self.project_dir)
                self.kill_files(submit_flags)
        elif modulename in RESUBMIT_MODULES:
            self.resubmit_module(modulename)
            # Project is restarted by previous call.
            return

        # Restart the project
        self.restart_project()

    @staticmethod
    def run(args):
        '''Main entry point'''
        tool = iBRAINTool(args.path.strip(), args.is_multiplate,
                          VERBOSITY_LEVELS[args.verbosity])
        if args.reset:
            tool.reset_project(args.force)
            return
        if args.resubmit:
            tool.resubmit(args.resubmit)
            return
        if args.restart:
            if args.restart != 'project':
                # print(args.restart)
                # if ',' in args.restart:
                #     modules = args.restart.split(',')
                # print(modules)
                # exit()
                tool.restart_module(args.restart)
            else:
                tool.restart_project()
            return
        raise iBRAINToolError('Nothing to do')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='iBRAINTool', description='Utility'
                                     'for simplified handling of iBRAIN '
                                     'project(s)')
    parser.add_argument('--version', action='version', version='%(prog)s v' +
                        __version__)
    parser.add_argument('--test', action='store_true',
                        help='Perform doctests')
    parser.add_argument("-v", "--verbosity", action="count", default=0,
                        help='Increase logging verbosity '
                        '(-v WARN, -vv INFO, -vvv DEBUG)')
    parser.add_argument('-d', '--dry-run', action='store_true',
                        default=False, help='DRY_RUN mode ("off" by '
                        'default)')
    parser.add_argument('--is-multiplate', action='store_true',
                        default=False, help='Is a multi-plate project?')
    parser.add_argument('-p', '--path', nargs='?', default=os.getcwd(),
                        help='Path to an iBRAIN project to work with')
    parser.add_argument('--force', action='store_true', default=False,
                        help='Force action')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-r', '--restart', choices=RESTART_CHOICES, nargs='?',
                       const='project', default='project',
                       help='Restart module or project. This kills .MAT '
                       'measurement files if found any. Use --resubmit '
                       'command if you want to resume as much as possible')
    group.add_argument('--reset', action='store_true', default=False,
                       help='Reset project by killing files and folders of '
                       'any output or result')
    group.add_argument('--resubmit', choices=RESUBMIT_CHOICES,
                       help='Resubmit module, trying to resume based on '
                       'previous results (default: batchjobs)')

    args = parser.parse_args()

    if args.dry_run:
        DRY_RUN = args.dry_run
    if not args.verbosity:
        args.verbosity

    if args.test:
        import doctest
        doctest.testmod()
        exit()

    try:
        iBRAINTool.run(args)
    except iBRAINToolError as error:
        print('Error: %s' % error.message, file=sys.stderr)



from brainy.scheduler.base import (SHORT_QUEUE, NORM_QUEUE, LONG_QUEUE,
                                   PENDING_STATE, RUNNING_STATE, DONE_STATE,
                                   JOB_STATES, BrainyScheduler)
import logging
logger = logging.getLogger(__name__)


class NoLsfSchedulerFound(Exception):
    '''No LSF scheduler'''


class Lsf(BrainyScheduler):

    def __init__(self):
        self.bsub = None
        self.bjobs = None
        self.__bjobs_cache = None
        self.bkill = None
        self.init_scheduling()
        self.states_map = {
            PENDING_STATE: 'PEND',
            RUNNING_STATE: 'RUN',
            DONE_STATE: 'DONE',
        }

    def init_scheduling(self):
        try:
            from sh import bjobs as _bjobs, bsub as _bsub, bkill as _bkill
            self.bsub = _bsub
            self.bjobs = _bjobs
            self.bkill = _bkill
        except ImportError:
            exception = NoLsfSchedulerFound('Failed to locate LSF commands '
                                            'like bjobs, bsub, bkill. Is LSF '
                                            'installed?')
            logger.warn(exception)

    def bjobs(self, *args, **kwds):
        if not self.__bjobs_cache:
            self.__bjobs_cache = self._bjobs(*args, **kwds)
        return self.__bjobs_cache

    def submit_job(self, shell_command, queue, report_file):
        '''Submit job using *bsub* command.'''
        try:
            self.bsub(
                '-W', queue,
                '-o', report_file,
                shell_command,
            )
        except Exception as error:
            logger.exception(error)
            return 'Failed to submit new job: %s' % shell_command

        logger.info('Submitting new job: %s', shell_command)
        logger.info('Report file will be written to: %s' % report_file)
        return ('Submitting new job: "%s"\n' +
                'Report file will be written to: %s') % \
               (shell_command, report_file)

    def count_working_jobs(self, key=None):
        '''
        Find out how many jobs are both PENDING and RUNNING. Require job
        description to contain the **key** substring. If key is None,
        then no filtering is done.
        '''
        jobs_list = self.bjobs('-aw').split('\n')
        if len(jobs_list) == 0:
            raise Exception('Failed to run bjobs')
        elif len(jobs_list) == 1 and 'No job found' in jobs_list:
            return 0
        working_jobs = [job_info for job_info in jobs_list
                        if ' RUN ' in job_info or ' PEND ' in job_info]
        logger.debug('Total number of working jobs found: %d' %
                     len(working_jobs))
        if key is None:
            # No filtering is applied.
            filtered_jobs = working_jobs
        else:
            # Sieve only matched job info strings.
            filtered_jobs = [job_info for job_info in jobs_list
                             if key in job_info]
        return len(filtered_jobs)

    def list_jobs(self, states):
        assert all([(state in JOB_STATES) for state in states])
        # TODO: make pure python implementation for state filtering.
        return self.bjobs('-aw').split('\n')



import logging
import atexit
from tree_output import HierarchicalOutput
from tree_output.log_handler import HierarchicalOutputHandler
LOGGING_OPTIONS = ('silent', 'console', 'json')


def make_console_handler(level):
    console = logging.StreamHandler()
    console.setLevel(level)
    return console


def make_hierarchical_handler(format):
    houtput = HierarchicalOutput.factory(format='json')
    return HierarchicalOutputHandler(houtput=houtput)


# We always need a json handler for project reports.
json_handler = make_hierarchical_handler('json')
json_handler.setLevel(logging.INFO)


def output_json():
    print str(json_handler.houtput)


def setup_logging(option, level=logging.INFO):
    if option not in LOGGING_OPTIONS:
        raise Exception('Unknown logging option: %s' % option)

    logging.root.setLevel(level)
    logging.root.addHandler(json_handler)

    # 'silent handler is ignored'
    if option == 'console':
        console_handler = make_console_handler(level)
        format_str = '%(asctime)s %(name)-30s %(levelname)-8s %(message)s'
        datefmt_str = '%m-%d %H:%M'
        console_handler.setFormatter(
            logging.Formatter(format_str, datefmt_str))
        logging.root.addHandler(console_handler)
    elif option == 'json':
        atexit.register(output_json)



import os
import shutil
import logging
from datetime import datetime
from findtools.find_files import (find_files, Match, MatchAnyPatternsAndTypes,
                                  MatchAllPatternsAndTypes)
from brainy.errors import BrainyProcessError
from brainy.process.code import PythonCodeProcess
from brainy.process.decorator import (
    format_with_params, require_keys_in_description,
    require_key_in_description)
logger = logging.getLogger(__name__)


KNOWN_MICROSCOPES = ['CV7K']


def get_timestamp_str():
    return datetime.now().strftime('%Y%m%d%H%M%S')


def backup_batch_folder(data_path, backups_path):
    '''Will recursively copy current BATCH to backups path'''
    # Make sure folders exist.
    if not os.path.exists(data_path):
        raise IOError('BATCH path was not found: %s' % data_path)
    if not os.path.exists(backups_path):
        raise IOError('BACKUPS path was not found: %s' % backups_path)

    # Making a backup copy.
    backuped_data_path = os.path.join(backups_path,
                                      'BATCH_%s' % get_timestamp_str())
    if os.path.exists(backuped_data_path):
        raise IOError('BACKUP destination already exists: %s' %
                      backuped_data_path)
    # Note: the destination directory must not already exist.
    shutil.copytree(data_path, backuped_data_path)


class BackupPreviousBatch(PythonCodeProcess):
    '''Backup BATCH of the successful RUN of the CellProfiller'''

    @property
    def backups_path(self):
        return self.restrict_to_safe_path(self.parameters.get(
            'backups_path',
            os.path.join(self.process_path, 'BACKUPS'),
        ))

    def put_on(self):
        super(BackupPreviousBatch, self).put_on()
        # Recreate missing BACKUPS folder path.
        if not os.path.exists(self.backups_path):
            os.makedirs(self.backups_path)

    def get_python_code(self):
        '''
        Note that the interpreter call is included by
        self.submit_python_code()
        '''
        return '''
        from brainy.pipes.Tools import backup_batch_folder

        backup_batch_folder('%(data_path)s', '%(backups_path)s')
        ''' % {
            'data_path': self.data_path,
            'backups_path': self.backups_path,
        }

    def has_data(self):
        '''
        If backups folder is empty, it means no backup was done.
        '''
        previous_backups = find_files(
            path=self.backups_path,
            match=Match(filetype='directory', name='BATCH_*'),
            recursive=False,
        )
        for item in previous_backups:
            return True
        return False


def move_microscope_metadata(tiff_path, metadata_path):
    # Make sure folders exist.
    if not os.path.exists(tiff_path):
        raise IOError('TIFF path was not found: %s' % tiff_path)
    if not os.path.exists(metadata_path):
        raise IOError('METADATA path was not found: %s' % metadata_path)

    # Roll over possible types.
    for microscope_type in KNOWN_MICROSCOPES:
        print '<!-- Checking if %s meta data is present -->' % microscope_type
        microscope_metadata_path = os.path.join(metadata_path, microscope_type)
        if microscope_type == 'CV7K':
            masks = [
                'geometry_parameter.xml',
                'MeasurementData.mlf',
                'MeasurementDetail.mrf',
                # e.g.: 1038402001_Greiner_#781091.wpp
                '*.wpp',
                # e.g.: 140314_InSituGFP.mes
                '*.mes',
                # e.g.: 140324-pilot-GFP-InSitu-gfp.wpi
                '*.wpi',
                # e.g.: DC_Andor #1_CAM1.tif
                r'/DC_Andor\ \#.*_CAM\d\.(tif|png)$/',
                # e.g.: SC_BP445-45_40x_M10_CH01.tif
                r'/SC_BP.*?CH\d*?\.(tif|png)$/',
            ]
            metadata_files = list(
                find_files(
                    path=tiff_path,
                    match=MatchAnyPatternsAndTypes(
                        filetypes=['f'],
                        names=masks,
                    ),
                )
            )
            if len(metadata_files) > 0:
                # Detected files for the microscope.
                if not os.path.exists(microscope_metadata_path):
                    os.mkdir(microscope_metadata_path)
                # Move files
                for metadata_file in metadata_files:
                    destination = os.path.join(
                        microscope_metadata_path,
                        os.path.basename(metadata_file),
                    )
                    print '<!-- Moving %s metadata: %s -&gt; %s -->' %\
                          (microscope_type, metadata_file, destination)
                    os.rename(metadata_file, destination)


def relative_symlink(source, target):
    if source[0] != '/' or target[0] != '/':
        raise ValueError('Arguments can be only absolute pathnames.')
    if source.startswith(target):
        raise ValueError('Source includes target pathname.')
    prefix = os.path.commonprefix([source, target])
    # Test for shared prefix to avoid bugs.
    if not prefix or prefix == '/':
        raise ValueError('No common shared prefix found.')
    if not prefix.endswith('/'):
        # Cut non-dir part.
        prefix = os.path.dirname(prefix)
    cut_index = len(prefix)
    relative_source = source[cut_index:]
    if relative_source.startswith('/'):
        relative_source = relative_source[1:]
    if len(relative_source) == 0:
        relative_source = prefix
    target_dir = target[cut_index:]
    # Find relative prefix.
    depth = len(target_dir.split(os.path.sep)) - 1
    relative_prefix = os.path.sep.join(['..' for foo in range(depth)])
    relative_source = os.path.sep.join((relative_prefix, relative_source))
    # Make missing target folders.
    parent_dir = os.path.dirname(target)
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)
    os.symlink(relative_source, target)


@require_keys_in_description('file_patterns')
class LinkFiles(PythonCodeProcess):
    '''
    Symlink or hardlink files using relative path to the current process_path.
    '''

    @property
    @format_with_params
    @require_key_in_description
    def source_location(self):
        pass

    @property
    @format_with_params
    @require_key_in_description
    def target_location(self):
        pass

    @property
    def file_type(self):
        '''By default, we link files, not folders. Use 'd' for folders.'''
        return self.description.get('file_type', 'f')

    @property
    def recursively(self):
        '''Recursively look for patterns in source_path'''
        return bool(self.description.get('recursively', False))

    def put_on(self):
        super(LinkFiles, self).put_on()
        # Create miss)ing process folder path.
        if not os.path.exists(self.data_path):
            os.makedirs(self.data_path)

    @staticmethod
    def link(source_path, target_path, patterns, link_type='hard',
             file_type='f', recursively=False):
        '''
        Expect keys 'hardlink' and 'symlink' keys in
        description['file_patterns']. If pattern string starts and ends with
        '/' then it is a regexp, otherwise it is fnmatch.
        '''
        assert os.path.exists(source_path)
        if not os.path.exists(target_path):
            os.makedirs(target_path)
        file_matches = find_files(
            path=source_path,
            match=MatchAnyPatternsAndTypes(
                # Can also accept symlinks sources, not just files or dirs.
                filetypes=[file_type, 'l'],
                names=patterns,
            ),
            recursive=recursively,
        )
        if link_type == 'hardlink' and file_type == 'f':
            make_link = os.link
        elif link_type == 'symlink':
            #make_link = os.symlink
            make_link = relative_symlink
        else:
            raise IOError('Unsupported link type: %s' % link_type)
        for source_file in file_matches:
            link_path = os.path.join(target_path,
                                     os.path.basename(source_file))
            try:
                print 'Linking "%s" -> "%s"' % (source_file, link_path)
                make_link(source_file, link_path)
            except (OSError, IOError) as error:
                if 'File exists' in str(error):
                    message = 'Skipping, since it looks like linking was '\
                        'already done. Maybe you are trying to re-run project'\
                        ' incorrectly. Make sure to clean previous results '\
                        'before retrying.'
                    print '%s\n(Error: %s)' % (message, str(error))
                else:
                    message = 'Unknown input-output error.'
                    raise BrainyProcessError(warning=message,
                                             output=str(error))

    @staticmethod
    def build_linking_args(source_location, target_location,
                           nested_file_patterns, file_type, recursively):
        for link_type in ['hardlink', 'symlink']:
            if link_type in nested_file_patterns:
                if type(nested_file_patterns[link_type]) != list \
                        or len(nested_file_patterns[link_type]) == 0:
                    raise BrainyProcessError(
                        message='LinkFiles process requires a non empty list '
                                'of file patterns which can be match to '
                                'files in source_location.'
                    )
                args = {
                    'source_location': source_location,
                    'target_location': target_location,
                    'file_patterns': nested_file_patterns[link_type],
                    'link_type': link_type,
                    'file_type': file_type,
                    'recursively': recursively,
                }
                yield args

    def get_python_code(self):
        '''
        Note that the interpreter call is included by
        self.submit_python_code()
        '''
        assert 'hardlink' in self.file_patterns \
            or 'symlink' in self.file_patterns

        code = 'from brainy.pipes.Tools import LinkFiles\n'

        for args in LinkFiles.build_linking_args(self.source_location,
                                                 self.target_location,
                                                 self.file_patterns,
                                                 self.file_type,
                                                 self.recursively):
            code += '''LinkFiles.link(
    '%(source_location)s',
    '%(target_location)s',
    %(file_patterns)s,
    link_type='%(link_type)s',
    file_type='%(file_type)s',
    recursively=%(recursively)r,
)
''' % args
        return code

    def has_data(self):
        '''
        Check if all files matching given patterns have been linked.
        '''
        if not os.path.exists(self.target_location):
            raise BrainyProcessError(
                message='Expected target folder is not '
                        'found: %s' % self.target_location,
                message_type='warning')

        def get_name(root, name):
            return name

        # raise Exception(self.source_location)

        if 'symlink' in self.file_patterns:
            patterns = self.file_patterns['symlink']
            source_matches = list(find_files(
                path=self.source_location,
                # Match any source type, i.e. we can link any source.
                match=MatchAnyPatternsAndTypes(names=patterns),
                recursive=self.recursively,
            ))
            target_matches = list(find_files(
                path=self.target_location,
                match=MatchAnyPatternsAndTypes(
                    filetypes=['l'],
                    names=patterns,
                ),
                recursive=self.recursively,
            ))
            if len(source_matches) != len(target_matches):
                return False
            source_matches = sorted(source_matches)
            target_matches = sorted(target_matches)
            for index in range(len(target_matches)):
                points_to = os.readlink(target_matches[index])
                while points_to.startswith('../'):
                    points_to = points_to[3:]
                if not source_matches[index].endswith(points_to):
                    return False

        if 'hardlink' in self.file_patterns:
            patterns = self.file_patterns['hardlink']
            for file_type in ['d', 'f']:
                source_matches = list(find_files(
                    path=self.source_location,
                    # Match any source type, i.e. we can link any source.
                    match=MatchAllPatternsAndTypes(
                       filetypes=[file_type],
                       names=patterns,
                    ),
                    collect=get_name,
                    recursive=self.recursively,
                ))
                target_matches = list(find_files(
                    path=self.target_location,
                    match=MatchAnyPatternsAndTypes(
                        filetypes=[file_type],
                        names=patterns,
                    ),
                    collect=get_name,
                    recursive=self.recursively,
                ))
                source_matches = sorted(source_matches)
                target_matches = sorted(target_matches)
                if source_matches != target_matches:
                    return False
        return True



import os
import shutil
import logging
from brainy.flags import FlagManager
from brainy.utils import Timer
from brainy.project.report import BrainyReporter, report_data
from brainy.pipes.base import PROCESS_NAMESPACE, BrainyPipe
from brainy.errors import BrainyPipeFailure, ProccessEndedIncomplete
logger = logging.getLogger(__name__)


class PipesManager(FlagManager):

    def __init__(self, project):
        self.project = project
        self.pipes_namespace = PROCESS_NAMESPACE
        self.pipes_folder_files = [
            os.path.join(self.project_path, filename)
            for filename in os.listdir(self.project_path)
        ]
        self.__flag_prefix = self.project_path
        self.__pipelines = None

    @property
    def scheduler(self):
        return self.project.scheduler

    @property
    def config(self):
        return self.project.config

    @property
    def project_path(self):
        return self.project.path

    def _get_flag_prefix(self):
        return self.__flag_prefix

    @property
    def pipe_extension(self):
        return self.project.config['brainy']['pipe_extension']

    def get_class(self, pipe_type):
        pipe_type = self.pipes_namespace + '.' + pipe_type
        module_name, class_name = pipe_type.rsplit('.', 1)
        module = __import__(module_name, {}, {}, [class_name])
        return getattr(module, class_name)

    @property
    def pipelines(self):
        if self.__pipelines is None:
            # Repopulate dictionary.
            logger.info('Discovering pipelines.')
            pipes = dict()
            for definition_filename in self.pipes_folder_files:
                if not definition_filename.endswith(self.pipe_extension):
                    continue
                pipe = BrainyPipe(self)
                pipe.parse_definition_file(definition_filename)
                cls = self.get_class(pipe.definition['type'])
                # Note that we pass itself as a pipes_manager
                pipes[pipe.definition['name']] = cls(self, pipe.definition)
            self.__pipelines = self.sort_pipelines(pipes)
        return self.__pipelines

    def sort_pipelines(self, pipes):
        '''Reorder, tolerating declared dependencies found in definitions'''
        after_dag = dict()
        before_dag = dict()
        for depended_pipename in pipes:
            pipe = pipes[depended_pipename]
            if 'after' in pipe.definition:
                dependends_on = pipe.definition['after']
                if dependends_on not in after_dag:
                    after_dag[dependends_on] = list()
                after_dag[dependends_on].append(depended_pipename)
            if 'before' in pipe.definition:
                dependends_on = pipe.definition['before']
                if dependends_on not in before_dag:
                    before_dag[dependends_on] = list()
                before_dag[dependends_on].append(depended_pipename)

        def resolve_dependecy(name_a, name_b):
            # After
            if name_a in after_dag:
                if name_b not in after_dag:
                    # Second argument has no "after" dependencies.
                    if name_b in after_dag[name_a]:
                        return -1
                else:
                    # Second argument has "after" dependencies.
                    if name_b in after_dag[name_a] \
                            and name_a in after_dag[name_b]:
                        raise Exception('Recursive dependencies')
                    if name_b in after_dag[name_a]:
                        return -1
            if name_b in after_dag:
                if name_a not in after_dag:
                    # First argument has no "after" dependencies.
                    if name_a in after_dag[name_b]:
                        return 1
                else:
                    # First argument has "after" dependencies.
                    if name_a in after_dag[name_b] \
                            and name_b in after_dag[name_a]:
                        raise Exception('Recursive dependencies')
                    if name_a in after_dag[name_b]:
                        return 1
            # Before
            if name_a in before_dag:
                if name_b not in before_dag:
                    # Second argument has no "before" dependencies.
                    if name_b in before_dag[name_a]:
                        return 1
                else:
                    # Second argument has "before" dependencies.
                    if name_b in before_dag[name_a] \
                            and name_a in before_dag[name_b]:
                        raise Exception('Recursive dependencies')
                    if name_b in before_dag[name_a]:
                        return 1
            if name_b in before_dag:
                if name_a not in before_dag:
                    # First argument has no "before" dependencies.
                    if name_a in before_dag[name_b]:
                        return -1
                else:
                    # First argument has "before" dependencies.
                    if name_a in before_dag[name_b] \
                            and name_b in before_dag[name_a]:
                        raise Exception('Recursive dependencies')
                    if name_a in before_dag[name_b]:
                        return -1
            return 0

        pipenames = list(pipes.keys())
        sorted_pipenames = sorted(pipenames, cmp=resolve_dependecy)
        result = list()
        for pipename in sorted_pipenames:
            result.append(pipes[pipename])
        return result

    def execute_pipeline(self, pipeline):
        '''
        Execute passed pipeline process within the context of this
        PipesModule.
        '''
        try:
            BrainyReporter.append_report_pipe(pipeline.name)
            pipeline.communicate({'input': '{}'})
        except ProccessEndedIncomplete as process_error:
            logger.error('%s Stopped on process called: %s' %
                         (str(process_error), process_error.process.name))
            pipeline.has_failed = True  # Halt further progress
        except BrainyPipeFailure as failure:
            # Errors are reported inside individual pipeline.
            logger.error('A pipeline has failed. We can not continue.')
            logger.exception(failure)
            pipeline.has_failed = True

    def process_pipelines(self):
        BrainyReporter.start_report()
        previous_pipeline = None
        for pipeline in self.pipelines:
            # Start by expecting failure free run.
            pipeline.has_failed = False

            # Check if current pipeline is dependent on previous one.
            depends_on_previous = False
            if previous_pipeline is not None:
                if 'before' in previous_pipeline.definition:
                    depends_on_previous = \
                        previous_pipeline.definition['before'] == pipeline.name
                elif 'after' in pipeline.definition:
                    depends_on_previous = \
                        previous_pipeline.name == pipeline.definition['after']

            if depends_on_previous and previous_pipeline.has_failed:
                logger.warn(('%s is skipped. Previous pipe that we depend on'
                            ' has failed or did not complete.') %
                            pipeline.name)
                # If previous pipeline we are dependent on has failed, then
                # mark pipeline as failed too to inform the next dependent
                # pipeline about the failure.
                pipeline.has_failed = True
                previous_pipeline = pipeline
                continue

            # Execute current pipeline.
            self.execute_pipeline(pipeline)

            # Remember as previous.
            previous_pipeline = pipeline

            if pipeline.has_failed:
                logger.warning('Pipeline {%s} has failed.' % pipeline.name)
            else:
                logger.info('Pipeline {%s} run without fatal errors.' %
                            pipeline.name)


        # Finalize the report.
        BrainyReporter.finalize_report()
        BrainyReporter.save_report(self.project.report_prefix_path)
        BrainyReporter.update_or_generate_static_html(
            self.project.report_folder_path)

    def clean_pipelines_output(self):
        for pipeline in self.pipelines:
            if os.path.exists(pipeline.output_path):
                logger.warn('Recursively clean/remove subfolder: %s' %
                            pipeline.output_path)
                shutil.rmtree(pipeline.output_path)

    def run(self, command):
        if not hasattr(self, command):
            logger.error('Pipes manager does not know command called: %s' %
                         command)
            return
        try:
            # Obtain method.
            method = getattr(self, command)
            # Time method execution.
            timer = Timer()
            with timer:
                method()
            logger.info('Finished running <%s>. It took about %d (s)' % (
                        command, timer.duration_in_seconds()))
            report_data['duration_in_seconds'] = timer.duration_in_seconds()
        except Exception as error:
            # logger.error(error)
            logger.exception(error)
